# ------------------------------------------------------------------------------

cdef class Sys:

    """System utilities."""

    @classmethod
    def getVersion(
        cls,
        devel: bool = False,
        date: bool = False,
        author: bool = False) -> tuple[int, int, int]:
        """Return PETSc version information.

        Not collective.

        Parameters
        ----------
        devel
            Additionally, return whether using an in-development version.
        date
            Additionally, return date information.
        author
            Additionally, return author information.

        Returns
        -------
        major : int
            Major version number.
        minor : int
            Minor version number.
        micro : int
            Micro (or patch) version number.

        See Also
        --------
        petsc.PetscGetVersion, petsc.PetscGetVersionNumber

        """
        cdef char cversion[256]
        cdef PetscInt major=0, minor=0, micro=0, release=0
        CHKERR(PetscGetVersion(cversion, sizeof(cversion)))
        CHKERR(PetscGetVersionNumber(&major, &minor, &micro, &release))
        out = version = (toInt(major), toInt(minor), toInt(micro))
        if devel or date or author:
            out = [version]
            if devel:
                out.append(not <bint>release)
            if date:
                vstr = bytes2str(cversion)
                if release != 0:
                    date = vstr.split(",", 1)[-1].strip()
                else:
                    date = vstr.split("Git Date:")[-1].strip()
                out.append(date)
            if author:
                author = bytes2str(PETSC_AUTHOR_INFO).split('\n')
                author = tuple([s.strip() for s in author if s])
                out.append(author)
        return tuple(out)

    @classmethod
    def getVersionInfo(cls) -> dict[str, bool | int | str]:
        """Return PETSc version information.

        Not collective.

        Returns
        -------
        info : dict
            Dictionary with version information.

        See Also
        --------
        petsc.PetscGetVersion, petsc.PetscGetVersionNumber

        """
        version, dev, date, author = cls.getVersion(True, True, True)
        return dict(major      = version[0],
                    minor      = version[1],
                    subminor   = version[2],
                    release    = not dev,
                    date       = date,
                    authorinfo = author)

    # --- xxx ---

    @classmethod
    def isInitialized(cls) -> bool:
        """Return whether PETSc has been initialized.

        Not collective.

        See Also
        --------
        isFinalized

        """
        return toBool(PetscInitializeCalled)

    @classmethod
    def isFinalized(cls) -> bool:
        """Return whether PETSc has been finalized.

        Not collective.

        See Also
        --------
        isInitialized

        """
        return toBool(PetscFinalizeCalled)

    # --- xxx ---

    @classmethod
    def getDefaultComm(cls) -> Comm:
        """Get the default MPI communicator used to create PETSc objects.

        Not collective.

        See Also
        --------
        setDefaultComm

        """
        cdef Comm comm = Comm()
        comm.comm = PETSC_COMM_DEFAULT
        return comm

    @classmethod
    def setDefaultComm(cls, comm: Comm | None) -> None:
        """Set the default MPI communicator used to create PETSc objects.

        Logically collective.

        Parameters
        ----------
        comm
            MPI communicator. If set to `None`, uses `COMM_WORLD`.

        See Also
        --------
        getDefaultComm

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_WORLD)
        if ccomm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        global PETSC_COMM_DEFAULT
        PETSC_COMM_DEFAULT = ccomm

    # --- xxx ---

    @classmethod
    def Print(
        cls,
        *args: Any,
        sep: str = ' ',
        end: str = '\n',
        comm: Comm | None = None,
        **kwargs: Any) -> None: # noqa: E129
        """Print output from the first processor of a communicator.

        Collective.

        Parameters
        ----------
        *args
            Positional arguments.
        sep
            String inserted between values, by default a space.
        end
            String appended after the last value, by default a newline.
        comm
            MPI communicator, defaults to `getDefaultComm`.
        **kwargs
            Keyword arguments.

        See Also
        --------
        petsc.PetscPrintf

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        if comm_rank(ccomm) == 0:
            if not args: args = ('',)
            format = ['%s', sep] * len(args)
            format[-1] = end
            message = ''.join(format) % args
        else:
            message = ''
        cdef const char *m = NULL
        message = str2bytes(message, &m)
        CHKERR(PetscPrintf(ccomm, '%s', m))

    @classmethod
    def syncPrint(
        cls,
        *args: Any,
        sep: str = ' ',
        end: str = '\n',
        flush: bool = False,
        comm: Comm | None = None,
        **kwargs: Any) -> None: # noqa: E129
        """Print synchronized output from several processors of a communicator.

        Not collective.

        Parameters
        ----------
        *args
            Positional arguments.
        sep
            String inserted between values, by default a space.
        end
            String appended after the last value, by default a newline.
        flush
            Whether to flush output with `syncFlush`.
        comm
            MPI communicator, defaults to `getDefaultComm`.
        **kwargs
            Keyword arguments.

        See Also
        --------
        petsc.PetscSynchronizedPrintf, petsc.PetscSynchronizedFlush

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        if not args: args = ('',)
        format = ['%s', sep] * len(args)
        format[-1] = end
        message = ''.join(format) % args
        cdef const char *m = NULL
        message = str2bytes(message, &m)
        CHKERR(PetscSynchronizedPrintf(ccomm, '%s', m))
        if flush: CHKERR(PetscSynchronizedFlush(ccomm, PETSC_STDOUT))

    @classmethod
    def syncFlush(cls, comm: Comm | None = None) -> None:
        """Flush output from previous `syncPrint` calls.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `getDefaultComm`.

        See Also
        --------
        petsc.PetscSynchronizedPrintf, petsc.PetscSynchronizedFlush

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        CHKERR(PetscSynchronizedFlush(ccomm, PETSC_STDOUT))

    # --- xxx ---

    @classmethod
    def splitOwnership(
        cls,
        size: int | tuple[int, int],
        bsize: int | None = None,
        comm: Comm | None = None) -> tuple[int, int]:
        """Given a global (or local) size determines a local (or global) size.

        Collective.

        Parameters
        ----------
        size
            Global size ``N`` or 2-tuple ``(n, N)`` with local and global
            sizes. Either of ``n`` or ``N`` (but not both) can be `None`.
        bsize
            Block size, defaults to ``1``.
        comm
            MPI communicator, defaults to `getDefaultComm`.

        Returns
        -------
        n : int
            The local size.
        N : int
            The global size.

        Notes
        -----
        The ``size`` argument corresponds to the full size of the
        vector. That is, an array with 10 blocks and a block size of 3 will
        have a ``size`` of 30, not 10.

        See Also
        --------
        petsc.PetscSplitOwnership

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscInt bs=0, n=0, N=0
        Sys_Sizes(size, bsize, &bs, &n, &N)
        if bs == PETSC_DECIDE: bs = 1
        if n > 0: n = n // bs
        if N > 0: N = N // bs
        CHKERR(PetscSplitOwnership(ccomm, &n, &N))
        n = n * bs
        N = N * bs
        return (toInt(n), toInt(N))

    @classmethod
    def sleep(cls, seconds: float = 1.0) -> None:
        """Sleep some number of seconds.

        Not collective.

        Parameters
        ----------
        seconds
            Time to sleep in seconds.

        See Also
        --------
        petsc.PetscSleep

        """
        cdef PetscReal s = asReal(seconds)
        CHKERR(PetscSleep(s))

    # --- xxx ---

    @classmethod
    def pushErrorHandler(cls, errhandler: str) -> None:
        """Set the current error handler.

        Logically collective.

        Parameters
        ----------
        errhandler
            The name of the error handler.

        See Also
        --------
        petsc.PetscPushErrorHandler

        """
        cdef PetscErrorHandlerFunction handler = NULL
        if errhandler == "python":
            handler = <PetscErrorHandlerFunction> PetscPythonErrorHandler
        elif errhandler == "debugger":
            handler = PetscAttachDebuggerErrorHandler
        elif errhandler == "emacs":
            handler = PetscEmacsClientErrorHandler
        elif errhandler == "traceback":
            handler = PetscTraceBackErrorHandler
        elif errhandler == "ignore":
            handler = PetscIgnoreErrorHandler
        elif errhandler == "mpiabort":
            handler = PetscMPIAbortErrorHandler
        elif errhandler == "abort":
            handler = PetscAbortErrorHandler
        else:
            raise ValueError(f"unknown error handler: {errhandler!r}")
        CHKERR(PetscPushErrorHandler(handler, NULL))

    @classmethod
    def popErrorHandler(cls) -> None:
        """Remove the current error handler.

        Logically collective.

        See Also
        --------
        petsc.PetscPopErrorHandler

        """
        CHKERR(PetscPopErrorHandler())

    @classmethod
    def popSignalHandler(cls) -> None:
        """Remove the current signal handler.

        Logically collective.

        See Also
        --------
        petsc.PetscPopSignalHandler

        """
        CHKERR(PetscPopSignalHandler())

    @classmethod
    def infoAllow(
        cls,
        flag: bool,
        filename: str | None = None,
        mode: str = "w") -> None:
        """Enables or disables PETSc info messages.

        Not collective.

        Parameters
        ----------
        flag
            Whether to enable info messages.
        filename
            Name of a file where to dump output.
        mode
            Write mode for file, by default ``"w"``.

        See Also
        --------
        petsc.PetscInfoAllow, petsc.PetscInfoSetFile

        """
        cdef PetscBool tval = PETSC_FALSE
        cdef const char *cfilename = NULL
        cdef const char *cmode = NULL
        if flag: tval = PETSC_TRUE
        CHKERR(PetscInfoAllow(tval))
        if filename is not None:
            filename = str2bytes(filename, &cfilename)
            mode = str2bytes(mode, &cmode)
            CHKERR(PetscInfoSetFile(cfilename, cmode))

    @classmethod
    def registerCitation(cls, citation: str) -> None:
        """Register BibTeX citation.

        Not collective.

        Parameters
        ----------
        citation
            The BibTex citation entry to register.

        See Also
        --------
        petsc.PetscCitationsRegister

        """
        if not citation: raise ValueError("empty citation")
        cdef const char *cit = NULL
        citation = str2bytes(citation, &cit)
        cdef PetscBool flag = get_citation(citation)
        CHKERR(PetscCitationsRegister(cit, &flag))
        set_citation(citation, toBool(flag))

    @classmethod
    def hasExternalPackage(cls, package: str) -> bool:
        """Return whether PETSc has support for external package.

        Not collective.

        Parameters
        ----------
        package
            The external package name.

        See Also
        --------
        petsc.PetscHasExternalPackage

        """
        cdef const char *cpackage = NULL
        package = str2bytes(package, &cpackage)
        cdef PetscBool has = PETSC_FALSE
        CHKERR(PetscHasExternalPackage(cpackage, &has))
        return toBool(has)


cdef dict citations_registry = {}

cdef PetscBool get_citation(object citation):
    cdef bint is_set = citations_registry.get(citation)
    return PETSC_TRUE if is_set else PETSC_FALSE

cdef set_citation(object citation, bint is_set):
    citations_registry[citation] = is_set

# ------------------------------------------------------------------------------



# ruff: noqa: UP008,UP031
from Cython.Compiler import Options
from Cython.Compiler import PyrexTypes
from Cython.Compiler.ExprNodes import TupleNode
from Cython.Compiler.Visitor import CythonTransform
from Cython.Compiler.StringEncoding import EncodedString
from Cython.Compiler.AutoDocTransforms import (
    ExpressionWriter as BaseExpressionWriter,
    AnnotationWriter as BaseAnnotationWriter,
)
from Cython.Compiler.Errors import error


class ExpressionWriter(BaseExpressionWriter):
    def visit_IndexNode(self, node):
        self.visit(node.base)
        self.put('[')
        if isinstance(node.index, TupleNode):
            if node.index.subexpr_nodes():
                self.emit_sequence(node.index)
            else:
                self.put('()')
        else:
            self.visit(node.index)
        self.put(']')

    if hasattr(BaseExpressionWriter, 'emit_string'):
        def visit_UnicodeNode(self, node):
            self.emit_string(node, '')


class AnnotationWriter(ExpressionWriter, BaseAnnotationWriter):
    pass


class EmbedSignature(CythonTransform):
    def __init__(self, context):
        super(EmbedSignature, self).__init__(context)
        self.class_name = None
        self.class_node = None

    def _select_format(self, embed, clinic):
        return embed

    def _fmt_expr(self, node):
        writer = ExpressionWriter()
        return writer.write(node)

    def _fmt_annotation(self, node):
        writer = AnnotationWriter()
        return writer.write(node)

    def _fmt_arg(self, arg):
        annotation = None
        if arg.is_self_arg:
            doc = self._select_format(arg.name, '$self')
        elif arg.is_type_arg:
            doc = self._select_format(arg.name, '$type')
        else:
            doc = arg.name
            if arg.type is PyrexTypes.py_object_type:
                annotation = None
            else:
                annotation = arg.type.declaration_code('', for_display=1)
                if arg.default and arg.default.is_none:
                    annotation += ' | None'
        if arg.annotation:
            annotation = self._fmt_annotation(arg.annotation)
        annotation = self._select_format(annotation, None)
        if annotation:
            doc = doc + (': %s' % annotation)
            if arg.default:
                default = self._fmt_expr(arg.default)
                doc = doc + (' = %s' % default)
        elif arg.default:
            default = self._fmt_expr(arg.default)
            doc = doc + ('=%s' % default)
        return doc

    def _fmt_star_arg(self, arg):
        arg_doc = arg.name
        if arg.annotation:
            annotation = self._fmt_annotation(arg.annotation)
            arg_doc = arg_doc + (': %s' % annotation)
        return arg_doc

    def _fmt_arglist(
        self,
        args,
        npoargs=0,
        npargs=0,
        pargs=None,
        nkargs=0,
        kargs=None,
        hide_self=False,
    ):
        arglist = []
        for arg in args:
            if not hide_self or not arg.entry.is_self_arg:
                arg_doc = self._fmt_arg(arg)
                arglist.append(arg_doc)
        if pargs:
            arg_doc = self._fmt_star_arg(pargs)
            arglist.insert(npargs + npoargs, '*%s' % arg_doc)
        elif nkargs:
            arglist.insert(npargs + npoargs, '*')
        if npoargs:
            arglist.insert(npoargs, '/')
        if kargs:
            arg_doc = self._fmt_star_arg(kargs)
            arglist.append('**%s' % arg_doc)
        return arglist

    def _fmt_ret_type(self, ret):
        if ret is PyrexTypes.py_object_type:
            return None
        return ret.declaration_code('', for_display=1)

    def _fmt_signature(
        self,
        node,
        cls_name,
        func_name,
        args,
        npoargs=0,
        npargs=0,
        pargs=None,
        nkargs=0,
        kargs=None,
        return_expr=None,
        return_type=None,
        hide_self=False,
    ):
        arglist = self._fmt_arglist(
            args, npoargs, npargs, pargs, nkargs, kargs, hide_self=hide_self
        )
        arglist_doc = ', '.join(arglist)
        func_doc = '%s(%s)' % (func_name, arglist_doc)
        if cls_name:
            namespace = self._select_format('%s.' % cls_name, '')
            func_doc = '%s%s' % (namespace, func_doc)
        ret_doc = None
        if return_expr:
            ret_doc = self._fmt_annotation(return_expr)
        elif return_type:
            ret_doc = self._fmt_ret_type(return_type)
        if ret_doc:
            docfmt = self._select_format('%s -> %s', '%s -> (%s)')
            func_doc = docfmt % (func_doc, ret_doc)
        else:
            if not func_doc.startswith('_') and not func_name.startswith('_'):
                error(
                    node.pos,
                    f'cyautodoc._fmt_signature: missing return type for {func_doc}',
                )
        return func_doc

    def _fmt_relative_position(self, pos):
        return 'Source code at ' + ':'.join(
            (str(pos[0].get_filenametable_entry()), str(pos[1]))
        )

    def _embed_signature(self, signature, pos, node_doc):
        pos = self._fmt_relative_position(pos)
        if node_doc:
            docfmt = self._select_format('%s\n%s\n%s', '%s\n--\n\n%s')
            return docfmt % (signature, node_doc, pos)
        docfmt = self._select_format('%s\n%s', '%s\n--\n\n%s')
        return docfmt % (signature, pos)

    def __call__(self, node):
        if not Options.docstrings:
            return node
        return super(EmbedSignature, self).__call__(node)

    def visit_ClassDefNode(self, node):
        oldname = self.class_name
        oldclass = self.class_node
        self.class_node = node
        try:
            # PyClassDefNode
            self.class_name = node.name
        except AttributeError:
            # CClassDefNode
            self.class_name = node.class_name
        self.visitchildren(node)
        self.class_name = oldname
        self.class_node = oldclass
        return node

    def visit_LambdaNode(self, node):
        # lambda expressions so not have signature or inner functions
        return node

    def visit_DefNode(self, node):
        if not self.current_directives['embedsignature']:
            return node

        hide_self = False
        if node.entry.is_special:
            return node
        class_name, func_name = self.class_name, node.name

        npoargs = getattr(node, 'num_posonly_args', 0)
        nkargs = getattr(node, 'num_kwonly_args', 0)
        npargs = len(node.args) - nkargs - npoargs
        signature = self._fmt_signature(
            node,
            class_name,
            func_name,
            node.args,
            npoargs,
            npargs,
            node.star_arg,
            nkargs,
            node.starstar_arg,
            return_expr=node.return_type_annotation,
            return_type=None,
            hide_self=hide_self,
        )
        if signature:
            doc_holder = node.entry

            if doc_holder.doc is not None:
                old_doc = doc_holder.doc
            elif getattr(node, 'py_func', None) is not None:
                old_doc = node.py_func.entry.doc
            else:
                old_doc = None
            new_doc = self._embed_signature(signature, node.pos, old_doc)
            doc_holder.doc = EncodedString(new_doc)
            if getattr(node, 'py_func', None) is not None:
                node.py_func.entry.doc = EncodedString(new_doc)
        return node

    def visit_CFuncDefNode(self, node):
        if not self.current_directives['embedsignature']:
            return node
        if not node.overridable:  # not cpdef FOO(...):
            return node

        signature = self._fmt_signature(
            node,
            self.class_name,
            node.declarator.base.name,
            node.declarator.args,
            return_type=node.return_type,
        )
        if signature:
            if node.entry.doc is not None:
                old_doc = node.entry.doc
            elif getattr(node, 'py_func', None) is not None:
                old_doc = node.py_func.entry.doc
            else:
                old_doc = None
            new_doc = self._embed_signature(signature, node.pos, old_doc)
            node.entry.doc = EncodedString(new_doc)
            py_func = getattr(node, 'py_func', None)
            if py_func is not None:
                py_func.entry.doc = EncodedString(new_doc)
        return node

    def visit_PropertyNode(self, node):
        if not self.current_directives['embedsignature']:
            return node

        entry = node.entry
        body = node.body
        prop_name = entry.name
        type_name = None
        if entry.visibility == 'public':
            # property synthesised from a cdef public attribute
            type_name = entry.type.declaration_code('', for_display=1)
            if not entry.type.is_pyobject:
                type_name = "'%s'" % type_name
            elif entry.type.is_extension_type:
                type_name = entry.type.module_name + '.' + type_name
        if type_name is None:
            for stat in body.stats:
                if stat.name != '__get__':
                    continue
                cls_name = self.class_name
                if cls_name:
                    namespace = self._select_format('%s.' % cls_name, '')
                    prop_name = '%s%s' % (namespace, prop_name)
                ret_annotation = stat.return_type_annotation
                if ret_annotation:
                    type_name = self._fmt_annotation(ret_annotation)
        if type_name is not None:
            signature = '%s: %s' % (prop_name, type_name)
            new_doc = self._embed_signature(signature, node.pos, entry.doc)
            entry.doc = EncodedString(new_doc)
        return node


# Monkeypatch AutoDocTransforms.EmbedSignature
try:
    from Cython.Compiler import AutoDocTransforms

    AutoDocTransforms.EmbedSignature = EmbedSignature
except Exception as exc:
    import logging

    logging.Logger(__name__).exception(exc)

# Monkeypatch Nodes.raise_utility_code
try:
    from Cython.Compiler.Nodes import raise_utility_code

    code = raise_utility_code.impl
    try:
        ipos = code.index('if (tb) {\n#if CYTHON_COMPILING_IN_PYPY\n')
    except ValueError:
        ipos = None
    else:
        raise_utility_code.impl = code[:ipos] + code[ipos:].replace(
            'CYTHON_COMPILING_IN_PYPY', '!CYTHON_FAST_THREAD_STATE', 1
        )
    del raise_utility_code, code, ipos
except Exception as exc:
    import logging

    logging.Logger(__name__).exception(exc)



# --------------------------------------------------------------------

cdef class Comm:
    """Communicator object.

    Predefined instances:

    `COMM_NULL`
        The *null* (or invalid) communicator.
    `COMM_SELF`
        The *self* communicator.
    `COMM_WORLD`
        The *world* communicator.

    See Also
    --------
    Sys.setDefaultComm, Sys.getDefaultComm

    """

    #

    def __cinit__(self, comm=None):
        self.comm = def_Comm(comm, MPI_COMM_NULL)
        self.isdup = 0
        if self.comm != MPI_COMM_NULL:
            self.base = comm
        else:
            self.base = None

    def __dealloc__(self):
        if self.isdup:
            CHKERR(PetscCommDEALLOC(&self.comm))
        self.comm = MPI_COMM_NULL
        self.isdup = 0
        self.base = None

    def __richcmp__(self, other, int op):
        if not isinstance(self,  Comm): return NotImplemented
        if not isinstance(other, Comm): return NotImplemented
        if op!=2 and op!=3: raise TypeError("only '==' and '!='")
        cdef Comm s = self
        cdef Comm o = other
        cdef int eq = (op == 2)
        cdef MPI_Comm comm1 = s.comm
        cdef MPI_Comm comm2 = o.comm
        cdef int flag = 0
        if comm1 != MPI_COMM_NULL and comm2 != MPI_COMM_NULL:
            CHKERR(<PetscErrorCode>MPI_Comm_compare(comm1, comm2, &flag))
            if eq: return (flag==<int>MPI_IDENT or  flag==<int>MPI_CONGRUENT)
            else:  return (flag!=<int>MPI_IDENT and flag!=<int>MPI_CONGRUENT)
        else:
            if eq: return (comm1 == comm2)
            else:  return (comm1 != comm2)

    def __bool__(self) -> bool:
        return self.comm != MPI_COMM_NULL

    #

    def destroy(self) -> None:
        """Destroy the communicator.

        Collective.

        See Also
        --------
        petsc.PetscCommDestroy

        """
        if self.comm == MPI_COMM_NULL: return
        if not self.isdup:
            raise ValueError("communicator not owned")
        CHKERR(PetscCommDestroy(&self.comm))
        self.comm = MPI_COMM_NULL
        self.isdup = 0
        self.base = None

    def duplicate(self) -> Self:
        """Duplicate the communicator.

        Collective.

        See Also
        --------
        petsc.PetscCommDuplicate

        """
        if self.comm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        cdef MPI_Comm newcomm = MPI_COMM_NULL
        CHKERR(PetscCommDuplicate(self.comm, &newcomm, NULL))
        cdef Comm comm = type(self)()
        comm.comm  = newcomm
        comm.isdup = 1
        comm.base = self.base
        return comm

    def getSize(self) -> int:
        """Return the number of processes in the communicator.

        Not collective.

        """
        if self.comm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        cdef int size=0
        CHKERRMPI(MPI_Comm_size(self.comm, &size))
        return size

    def getRank(self) -> int:
        """Return the rank of the calling processes in the communicator.

        Not collective.

        """
        if self.comm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        cdef int rank=0
        CHKERRMPI(MPI_Comm_rank(self.comm, &rank))
        return rank

    def barrier(self) -> None:
        """Barrier synchronization.

        Collective.

        """
        if self.comm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        CHKERRMPI(MPI_Barrier(self.comm))

    # --- properties ---

    property size:
        """Communicator size."""
        def __get__(self) -> int:
            return self.getSize()

    property rank:
        """Communicator rank."""
        def __get__(self) -> int:
            return self.getRank()

    # --- Fortran support ---

    property fortran:
        """Fortran handle."""
        def __get__(self) -> int:
            cdef MPI_Comm comm = self.comm
            return MPI_Comm_c2f(comm)

    # --- mpi4py support ---

    def tompi4py(self) -> Intracomm:
        """Convert communicator to `mpi4py`.

        Not collective.

        See Also
        --------
        mpi4py.MPI.Comm, mpi4py.MPI.Intracomm

        """
        cdef MPI_Comm comm = self.comm
        return mpi4py_Comm_New(comm)

    # --- mpi4py compatibility API ---

    Free     = destroy
    Clone    = duplicate
    Dup      = duplicate
    Get_size = getSize
    Get_rank = getRank
    Barrier  = barrier

# --------------------------------------------------------------------

cdef Comm __COMM_NULL__  = Comm()
cdef Comm __COMM_SELF__  = Comm()
cdef Comm __COMM_WORLD__ = Comm()

COMM_NULL  = __COMM_NULL__
COMM_SELF  = __COMM_SELF__
COMM_WORLD = __COMM_WORLD__

# --------------------------------------------------------------------

cdef MPI_Comm PETSC_COMM_DEFAULT = MPI_COMM_NULL

cdef MPI_Comm GetComm(
    object comm, MPI_Comm defv,
) except? MPI_COMM_NULL:
    return def_Comm(comm, defv)

cdef MPI_Comm GetCommDefault():
    return PETSC_COMM_DEFAULT

# --------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef const char* PetscRandomType
    PetscRandomType PETSCRAND
    PetscRandomType PETSCRAND48
    PetscRandomType PETSCSPRNG
    PetscRandomType PETSCRANDER48
    PetscRandomType PETSCRANDOM123

    PetscErrorCode PetscRandomCreate(MPI_Comm, PetscRandom*)
    PetscErrorCode PetscRandomDestroy(PetscRandom*)
    PetscErrorCode PetscRandomView(PetscRandom, PetscViewer)

    PetscErrorCode PetscRandomSetType(PetscRandom, PetscRandomType)
    PetscErrorCode PetscRandomGetType(PetscRandom, PetscRandomType*)
    PetscErrorCode PetscRandomSetFromOptions(PetscRandom)

    PetscErrorCode PetscRandomGetValue(PetscRandom, PetscScalar*)
    PetscErrorCode PetscRandomGetValueReal(PetscRandom, PetscReal*)
    PetscErrorCode PetscRandomGetValueImaginary(PetscRandom, PetscScalar*)
    PetscErrorCode PetscRandomGetInterval(PetscRandom, PetscScalar*, PetscScalar*)
    PetscErrorCode PetscRandomSetInterval(PetscRandom, PetscScalar, PetscScalar)
    PetscErrorCode PetscRandomSetSeed(PetscRandom, PetscInt64)
    PetscErrorCode PetscRandomGetSeed(PetscRandom, PetscInt64*)
    PetscErrorCode PetscRandomSeed(PetscRandom)



# --------------------------------------------------------------------

class FEType(object):
    """The finite element types."""
    BASIC     = S_(PETSCFEBASIC)
    OPENCL    = S_(PETSCFEOPENCL)
    COMPOSITE = S_(PETSCFECOMPOSITE)

# --------------------------------------------------------------------


cdef class FE(Object):
    """A PETSc object that manages a finite element space."""

    Type = FEType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.fe
        self.fe = NULL

    def view(self, Viewer viewer=None) -> None:
        """View a `FE` object.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the graph.

        See Also
        --------
        petsc.PetscFEView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscFEView(self.fe, vwr))

    def destroy(self) -> Self:
        """Destroy the `FE` object.

        Collective.

        See Also
        --------
        petsc.PetscFEDestroy

        """
        CHKERR(PetscFEDestroy(&self.fe))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Create an empty `FE` object.

        Collective.

        The type can then be set with `setType`.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        setType, petsc.PetscFECreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscFE newfe = NULL
        CHKERR(PetscFECreate(ccomm, &newfe))
        CHKERR(PetscCLEAR(self.obj)); self.fe = newfe
        return self

    def createDefault(
        self,
        dim: int,
        nc: int,
        isSimplex: bool,
        qorder: int = DETERMINE,
        prefix: str | None = None,
        comm: Comm | None = None) -> Self:
        """Create a `FE` for basic FEM computation.

        Collective.

        Parameters
        ----------
        dim
            The spatial dimension.
        nc
            The number of components.
        isSimplex
            Flag for simplex reference cell, otherwise it's a tensor product.
        qorder
            The quadrature order or `DETERMINE` to use `Space` polynomial
            degree.
        prefix
            The options prefix, or `None`.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscFECreateDefault

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscFE newfe = NULL
        cdef PetscInt cdim = asInt(dim)
        cdef PetscInt cnc = asInt(nc)
        cdef PetscInt cqorder = asInt(qorder)
        cdef PetscBool cisSimplex = asBool(isSimplex)
        cdef const char *cprefix = NULL
        if prefix:
            prefix = str2bytes(prefix, &cprefix)
        CHKERR(PetscFECreateDefault(ccomm, cdim, cnc, cisSimplex, cprefix, cqorder, &newfe))
        CHKERR(PetscCLEAR(self.obj)); self.fe = newfe
        return self

    def createByCell(
        self,
        dim: int,
        nc: int,
        ctype: DM.PolytopeType,
        qorder: int = DETERMINE,
        prefix: str | None = None,
        comm: Comm | None = None) -> Self:
        """Create a `FE` for basic FEM computation.

        Collective.

        Parameters
        ----------
        dim
            The spatial dimension.
        nc
            The number of components.
        ctype
            The cell type.
        qorder
            The quadrature order or `DETERMINE` to use `Space` polynomial
            degree.
        prefix
            The options prefix, or `None`.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscFECreateByCell

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_SELF)
        cdef PetscFE newfe = NULL
        cdef PetscInt cdim = asInt(dim)
        cdef PetscInt cnc = asInt(nc)
        cdef PetscInt cqorder = asInt(qorder)
        cdef PetscDMPolytopeType cCellType = ctype
        cdef const char *cprefix = NULL
        if prefix:
            prefix = str2bytes(prefix, &cprefix)
        CHKERR(PetscFECreateDefault(ccomm, cdim, cnc, cCellType, cprefix, cqorder, &newfe))
        CHKERR(PetscCLEAR(self.obj)); self.fe = newfe
        return self

    def createLagrange(
        self,
        dim: int,
        nc: int,
        isSimplex: bool,
        k: int,
        qorder: int = DETERMINE,
        comm: Comm | None = None) -> Self:
        """Create a `FE` for the basic Lagrange space of degree k.

        Collective.

        Parameters
        ----------
        dim
            The spatial dimension.
        nc
            The number of components.
        isSimplex
            Flag for simplex reference cell, otherwise it's a tensor product.
        k
            The degree of the space.
        qorder
            The quadrature order or `DETERMINE` to use `Space` polynomial
            degree.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscFECreateLagrange

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscFE newfe = NULL
        cdef PetscInt cdim = asInt(dim)
        cdef PetscInt cnc = asInt(nc)
        cdef PetscInt ck = asInt(k)
        cdef PetscInt cqorder = asInt(qorder)
        cdef PetscBool cisSimplex = asBool(isSimplex)
        CHKERR(PetscFECreateLagrange(ccomm, cdim, cnc, cisSimplex, ck, cqorder, &newfe))
        CHKERR(PetscCLEAR(self.obj)); self.fe = newfe
        return self

    def getQuadrature(self) -> Quad:
        """Return the `Quad` used to calculate inner products.

        Not collective.

        See Also
        --------
        setQuadrature, petsc.PetscFEGetQuadrature

        """
        cdef Quad quad = Quad()
        CHKERR(PetscFEGetQuadrature(self.fe, &quad.quad))
        CHKERR(PetscINCREF(quad.obj))
        return quad

    def getDimension(self) -> int:
        """Return the dimension of the finite element space on a cell.

        Not collective.

        See Also
        --------
        petsc.PetscFEGetDimension

        """
        cdef PetscInt cdim = 0
        CHKERR(PetscFEGetDimension(self.fe, &cdim))
        return toInt(cdim)

    def getSpatialDimension(self) -> int:
        """Return the spatial dimension of the element.

        Not collective.

        See Also
        --------
        petsc.PetscFEGetSpatialDimension

        """
        cdef PetscInt csdim = 0
        CHKERR(PetscFEGetSpatialDimension(self.fe, &csdim))
        return toInt(csdim)

    def getNumComponents(self) -> int:
        """Return the number of components in the element.

        Not collective.

        See Also
        --------
        setNumComponents, petsc.PetscFEGetNumComponents

        """
        cdef PetscInt comp = 0
        CHKERR(PetscFEGetNumComponents(self.fe, &comp))
        return toInt(comp)

    def setNumComponents(self, comp: int) -> None:
        """Set the number of field components in the element.

        Not collective.

        Parameters
        ----------
        comp
            The number of field components.

        See Also
        --------
        getNumComponents, petsc.PetscFESetNumComponents

        """
        cdef PetscInt ccomp = asInt(comp)
        CHKERR(PetscFESetNumComponents(self.fe, ccomp))

    def getNumDof(self) -> ndarray:
        """Return the number of DOFs.

        Not collective.

        Return the number of DOFs (dual basis vectors) associated with mesh
        points on the reference cell of a given dimension.

        See Also
        --------
        petsc.PetscFEGetNumDof

        """
        cdef const PetscInt *numDof = NULL
        cdef PetscInt cdim = 0
        CHKERR(PetscFEGetDimension(self.fe, &cdim))
        CHKERR(PetscFEGetNumDof(self.fe, &numDof))
        return array_i(cdim, numDof)

    def getTileSizes(self) -> tuple[int, int, int, int]:
        """Return the tile sizes for evaluation.

        Not collective.

        Returns
        -------
        blockSize : int
            The number of elements in a block.
        numBlocks : int
            The number of blocks in a batch.
        batchSize : int
            The number of elements in a batch.
        numBatches : int
            The number of batches in a chunk.

        See Also
        --------
        setTileSizes, petsc.PetscFEGetTileSizes

        """
        cdef PetscInt blockSize = 0, numBlocks = 0
        cdef PetscInt batchSize = 0, numBatches = 0
        CHKERR(PetscFEGetTileSizes(self.fe, &blockSize, &numBlocks, &batchSize, &numBatches))
        return toInt(blockSize), toInt(numBlocks), toInt(batchSize), toInt(numBatches)

    def setTileSizes(
        self,
        blockSize: int,
        numBlocks: int,
        batchSize: int,
        numBatches: int) -> None:
        """Set the tile sizes for evaluation.

        Not collective.

        Parameters
        ----------
        blockSize
            The number of elements in a block.
        numBlocks
            The number of blocks in a batch.
        batchSize
            The number of elements in a batch.
        numBatches
            The number of batches in a chunk.

        See Also
        --------
        getTileSizes, petsc.PetscFESetTileSizes

        """
        cdef PetscInt cblockSize = asInt(blockSize), cnumBlocks = asInt(numBlocks)
        cdef PetscInt cbatchSize = asInt(batchSize), cnumBatches = asInt(numBatches)
        CHKERR(PetscFESetTileSizes(self.fe, cblockSize, cnumBlocks, cbatchSize, cnumBatches))

    def getFaceQuadrature(self) -> Quad:
        """Return the `Quad` used to calculate inner products on faces.

        Not collective.

        See Also
        --------
        setFaceQuadrature, petsc.PetscFEGetFaceQuadrature

        """
        cdef Quad quad = Quad()
        CHKERR(PetscFEGetFaceQuadrature(self.fe, &quad.quad))
        CHKERR(PetscINCREF(quad.obj))
        return quad

    def setQuadrature(self, Quad quad) -> Self:
        """Set the `Quad` used to calculate inner products.

        Not collective.

        Parameters
        ----------
        quad
            The `Quad` object.

        See Also
        --------
        getQuadrature, petsc.PetscFESetQuadrature

        """
        CHKERR(PetscFESetQuadrature(self.fe, quad.quad))
        return self

    def setFaceQuadrature(self, Quad quad) -> Quad:
        """Set the `Quad` used to calculate inner products on faces.

        Not collective.

        Parameters
        ----------
        quad
            The `Quad` object.

        See Also
        --------
        getFaceQuadrature, petsc.PetscFESetFaceQuadrature

        """
        CHKERR(PetscFESetFaceQuadrature(self.fe, quad.quad))
        return self

    def setType(self, fe_type: Type | str) -> Self:
        """Build a particular `FE`.

        Collective.

        Parameters
        ----------
        fe_type
            The kind of FEM space.

        See Also
        --------
        petsc.PetscFESetType

        """
        cdef PetscFEType cval = NULL
        fe_type = str2bytes(fe_type, &cval)
        CHKERR(PetscFESetType(self.fe, cval))
        return self

    def getBasisSpace(self) -> Space:
        """Return the `Space` used for the approximation of the `FE` solution.

        Not collective.

        See Also
        --------
        setBasisSpace, petsc.PetscFEGetBasisSpace

        """
        cdef Space sp = Space()
        CHKERR(PetscFEGetBasisSpace(self.fe, &sp.space))
        CHKERR(PetscINCREF(sp.obj))
        return sp

    def setBasisSpace(self, Space sp) -> None:
        """Set the `Space` used for the approximation of the solution.

        Not collective.

        Parameters
        ----------
        sp
            The `Space` object.

        See Also
        --------
        getBasisSpace, petsc.PetscFESetBasisSpace

        """
        CHKERR(PetscFESetBasisSpace(self.fe, sp.space))

    def setFromOptions(self) -> None:
        """Set parameters in a `FE` from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.PetscFESetFromOptions

        """
        CHKERR(PetscFESetFromOptions(self.fe))

    def setUp(self) -> None:
        """Construct data structures for the `FE` after the `Type` has been set.

        Collective.

        See Also
        --------
        petsc.PetscFESetUp

        """
        CHKERR(PetscFESetUp(self.fe))

    def getDualSpace(self) -> DualSpace:
        """Return the `DualSpace` used to define the inner product for the `FE`.

        Not collective.

        See Also
        --------
        setDualSpace, DualSpace, petsc.PetscFEGetDualSpace

        """
        cdef DualSpace dspace = DualSpace()
        CHKERR(PetscFEGetDualSpace(self.fe, &dspace.dualspace))
        CHKERR(PetscINCREF(dspace.obj))
        return dspace

    def setDualSpace(self, DualSpace dspace) -> None:
        """Set the `DualSpace` used to define the inner product.

        Not collective.

        Parameters
        ----------
        dspace
            The `DualSpace` object.

        See Also
        --------
        getDualSpace, DualSpace, petsc.PetscFESetDualSpace

        """
        CHKERR(PetscFESetDualSpace(self.fe, dspace.dualspace))

# --------------------------------------------------------------------

del FEType

# --------------------------------------------------------------------



from petsc4py import PETSc
import unittest
import numpy as np


class TestDMShell(unittest.TestCase):
    COMM = PETSc.COMM_WORLD

    def setUp(self):
        self.dm = PETSc.DMShell().create(comm=self.COMM)

    def tearDown(self):
        self.dm.destroy()
        self.dm = None
        PETSc.garbage_cleanup()

    def testSetGlobalVector(self):
        vec = PETSc.Vec().create(comm=self.COMM)
        vec.setSizes((10, None))
        vec.setUp()
        self.dm.setGlobalVector(vec)
        gvec = self.dm.createGlobalVector()
        self.assertEqual(vec.getSizes(), gvec.getSizes())
        self.assertEqual(vec.comm, gvec.comm)

    def testSetCreateGlobalVector(self):
        def create_vec(dm):
            v = PETSc.Vec().create(comm=dm.comm)
            v.setSizes((10, None))
            v.setUp()
            return v

        self.dm.setCreateGlobalVector(create_vec)
        gvec = self.dm.createGlobalVector()
        self.assertEqual(gvec.comm, self.dm.comm)
        self.assertEqual(gvec.getLocalSize(), 10)

    def testSetLocalVector(self):
        vec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        vec.setSizes((1 + 10 * self.COMM.rank, None))
        vec.setUp()
        self.dm.setLocalVector(vec)
        lvec = self.dm.createLocalVector()
        self.assertEqual(vec.getSizes(), lvec.getSizes())
        lsize, gsize = lvec.getSizes()
        self.assertEqual(lsize, gsize)
        self.assertEqual(lvec.comm, PETSc.COMM_SELF)

    def testSetCreateLocalVector(self):
        def create_vec(dm):
            v = PETSc.Vec().create(comm=PETSc.COMM_SELF)
            v.setSizes((1 + 10 * dm.comm.rank, None))
            v.setUp()
            return v

        self.dm.setCreateLocalVector(create_vec)
        lvec = self.dm.createLocalVector()
        lsize, gsize = lvec.getSizes()
        self.assertEqual(lsize, gsize)
        self.assertEqual(lsize, 1 + 10 * self.dm.comm.rank)
        self.assertEqual(lvec.comm, PETSc.COMM_SELF)

    def testSetMatrix(self):
        mat = PETSc.Mat().create(comm=self.COMM)
        mat.setSizes(((10, None), (2, None)))
        mat.setUp()
        mat.assemble()
        self.dm.setMatrix(mat)
        nmat = self.dm.createMatrix()
        self.assertEqual(nmat.getSizes(), mat.getSizes())

    def testSetCreateMatrix(self):
        def create_mat(dm):
            mat = PETSc.Mat().create(comm=self.COMM)
            mat.setSizes(((10, None), (2, None)))
            mat.setUp()
            return mat

        self.dm.setCreateMatrix(create_mat)
        nmat = self.dm.createMatrix()
        self.assertEqual(nmat.getSizes(), create_mat(self.dm).getSizes())

    def testSetCreateFieldDecomposition(self):
        def create_field_decomposition_only_is(dm):
            return None, [PETSc.IS().createStride(1) for _ in range(2)], None

        def create_field_decomposition_only_is_names(dm):
            _, ises, _ = create_field_decomposition_only_is(dm)
            names = [f'f_{i}' for i in range(len(ises))]
            return names, ises, None

        def create_field_decomposition_only_is_dms(dm):
            _, ises, _ = create_field_decomposition_only_is(dm)
            dms = [dm.clone() for _ in range(len(ises))]
            return _, ises, dms

        def create_field_decomposition_only_full(dm):
            names, ises, _ = create_field_decomposition_only_is_names(dm)
            dms = [dm.clone() for _ in range(len(ises))]
            return names, ises, dms

        tests = [
            (create_field_decomposition_only_is, False, False),
            (create_field_decomposition_only_is_names, True, False),
            (create_field_decomposition_only_is_dms, False, True),
            (create_field_decomposition_only_full, True, True),
        ]
        for test_f, has_names, has_dms in tests:
            self.dm.setCreateFieldDecomposition(test_f)
            names, ises, dms = self.dm.createFieldDecomposition()
            self.assertEqual(len(names), len(ises))
            self.assertEqual(len(dms), len(ises))
            if has_names:
                checknames = [f'f_{i}' for i in range(len(ises))]
                self.assertEqual(names, checknames)
            else:
                for n in names:
                    self.assertTrue(n is None)
            for dm in dms:
                if has_dms:
                    self.assertEqual(type(dm), PETSc.DM)
                else:
                    self.assertTrue(dm is None)

    def testGlobalToLocal(self):
        def begin(dm, ivec, mode, ovec):
            if mode == PETSc.InsertMode.INSERT_VALUES:
                ovec[...] = ivec[...]
            elif mode == PETSc.InsertMode.ADD_VALUES:
                ovec[...] += ivec[...]

        def end(dm, ivec, mode, ovec):
            pass

        vec = PETSc.Vec().create(comm=self.COMM)
        vec.setSizes((10, None))
        vec.setUp()
        vec[...] = self.dm.comm.rank + 1
        ovec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        ovec.setSizes((10, None))
        ovec.setUp()
        self.dm.setGlobalToLocal(begin, end)
        self.dm.globalToLocal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)
        self.assertTrue(np.allclose(vec.getArray(), ovec.getArray()))
        self.dm.globalToLocal(vec, ovec, addv=PETSc.InsertMode.ADD_VALUES)
        self.assertTrue(np.allclose(2 * vec.getArray(), ovec.getArray()))

    def testLocalToGlobal(self):
        def begin(dm, ivec, mode, ovec):
            if mode == PETSc.InsertMode.INSERT_VALUES:
                ovec[...] = ivec[...]
            elif mode == PETSc.InsertMode.ADD_VALUES:
                ovec[...] += ivec[...]

        def end(dm, ivec, mode, ovec):
            pass

        vec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        vec.setSizes((10, None))
        vec.setUp()
        vec[...] = self.dm.comm.rank + 1
        ovec = PETSc.Vec().create(comm=self.COMM)
        ovec.setSizes((10, None))
        ovec.setUp()
        self.dm.setLocalToGlobal(begin, end)
        self.dm.localToGlobal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)
        self.assertTrue(np.allclose(vec.getArray(), ovec.getArray()))
        self.dm.localToGlobal(vec, ovec, addv=PETSc.InsertMode.ADD_VALUES)
        self.assertTrue(np.allclose(2 * vec.getArray(), ovec.getArray()))

    def testLocalToLocal(self):
        def begin(dm, ivec, mode, ovec):
            if mode == PETSc.InsertMode.INSERT_VALUES:
                ovec[...] = ivec[...]
            elif mode == PETSc.InsertMode.ADD_VALUES:
                ovec[...] += ivec[...]

        def end(dm, ivec, mode, ovec):
            pass

        vec = PETSc.Vec().create(comm=PETSc.COMM_SELF)
        vec.setSizes((10, None))
        vec.setUp()
        vec[...] = self.dm.comm.rank + 1
        ovec = vec.duplicate()
        self.dm.setLocalToLocal(begin, end)
        self.dm.localToLocal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)
        self.assertTrue(np.allclose(vec.getArray(), ovec.getArray()))
        self.dm.localToLocal(vec, ovec, addv=PETSc.InsertMode.ADD_VALUES)
        self.assertTrue(np.allclose(2 * vec.getArray(), ovec.getArray()))

    def testGlobalToLocalVecScatter(self):
        vec = PETSc.Vec().create()
        vec.setSizes((10, None))
        vec.setUp()
        sct, ovec = PETSc.Scatter.toAll(vec)
        self.dm.setGlobalToLocalVecScatter(sct)
        self.dm.globalToLocal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)

    def testLocalToGlobalVecScatter(self):
        vec = PETSc.Vec().create()
        vec.setSizes((10, None))
        vec.setUp()
        sct, ovec = PETSc.Scatter.toAll(vec)
        self.dm.setLocalToGlobalVecScatter(sct)
        self.dm.localToGlobal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)

    def testLocalToLocalVecScatter(self):
        vec = PETSc.Vec().create()
        vec.setSizes((10, None))
        vec.setUp()
        sct, ovec = PETSc.Scatter.toAll(vec)
        self.dm.setLocalToLocalVecScatter(sct)
        self.dm.localToLocal(vec, ovec, addv=PETSc.InsertMode.INSERT_VALUES)

    def testCoarsenRefine(self):
        cdm = PETSc.DMShell().create(comm=self.COMM)

        def coarsen(dm, comm):
            return cdm

        def refine(dm, comm):
            return self.dm

        cdm.setRefine(refine)
        self.dm.setCoarsen(coarsen)
        coarsened = self.dm.coarsen()
        self.assertEqual(coarsened, cdm)
        refined = coarsened.refine()
        self.assertEqual(refined, self.dm)

    def testCreateInterpolation(self):
        mat = PETSc.Mat().create()
        mat.setSizes(((10, None), (10, None)))
        mat.setUp()
        vec = PETSc.Vec().create()
        vec.setSizes((10, None))
        vec.setUp()

        def create_interp(dm, dmf):
            return mat, vec

        self.dm.setCreateInterpolation(create_interp)
        m, v = self.dm.createInterpolation(self.dm)
        self.assertEqual(m, mat)
        self.assertEqual(v, vec)

    def testCreateInjection(self):
        mat = PETSc.Mat().create()
        mat.setSizes(((10, None), (10, None)))
        mat.setUp()

        def create_inject(dm, dmf):
            return mat

        self.dm.setCreateInjection(create_inject)
        m = self.dm.createInjection(self.dm)
        self.assertEqual(m, mat)


if __name__ == '__main__':
    unittest.main()



cdef extern from * nogil:

    ctypedef const char* PetscTAOType "TaoType"
    PetscTAOType TAOLMVM
    PetscTAOType TAONLS
    PetscTAOType TAONTR
    PetscTAOType TAONTL
    PetscTAOType TAOCG
    PetscTAOType TAOTRON
    PetscTAOType TAOOWLQN
    PetscTAOType TAOBMRM
    PetscTAOType TAOBLMVM
    PetscTAOType TAOBQNLS
    PetscTAOType TAOBNCG
    PetscTAOType TAOBNLS
    PetscTAOType TAOBNTR
    PetscTAOType TAOBNTL
    PetscTAOType TAOBQNKLS
    PetscTAOType TAOBQNKTR
    PetscTAOType TAOBQNKTL
    PetscTAOType TAOBQPIP
    PetscTAOType TAOGPCG
    PetscTAOType TAONM
    PetscTAOType TAOPOUNDERS
    PetscTAOType TAOBRGN
    PetscTAOType TAOLCL
    PetscTAOType TAOSSILS
    PetscTAOType TAOSSFLS
    PetscTAOType TAOASILS
    PetscTAOType TAOASFLS
    PetscTAOType TAOIPM
    PetscTAOType TAOPDIPM
    PetscTAOType TAOSHELL
    PetscTAOType TAOADMM
    PetscTAOType TAOALMM
    PetscTAOType TAOPYTHON

    ctypedef enum PetscTAOConvergedReason "TaoConvergedReason":
        # iterating
        TAO_CONTINUE_ITERATING
        # converged
        TAO_CONVERGED_GATOL
        TAO_CONVERGED_GRTOL
        TAO_CONVERGED_GTTOL
        TAO_CONVERGED_STEPTOL
        TAO_CONVERGED_MINF
        TAO_CONVERGED_USER
        # diverged
        TAO_DIVERGED_MAXITS
        TAO_DIVERGED_NAN
        TAO_DIVERGED_MAXFCN
        TAO_DIVERGED_LS_FAILURE
        TAO_DIVERGED_TR_REDUCTION
        TAO_DIVERGED_USER

    ctypedef PetscErrorCode (*PetscTaoMonitorDestroy)(void**)
    ctypedef PetscErrorCode PetscTaoConvergenceTest(PetscTAO, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoMonitor(PetscTAO, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoObjective(PetscTAO, PetscVec, PetscReal*, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoResidual(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoGradient(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoObjGrad(PetscTAO, PetscVec, PetscReal*, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoRegularizerObjGrad(PetscTAO, PetscVec, PetscReal*, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoVarBounds(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoConstraints(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoEqualityConstraints(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoInequalityConstraints(PetscTAO, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoHessian(PetscTAO, PetscVec, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoRegularizerHessian(PetscTAO, PetscVec, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobian(PetscTAO, PetscVec, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobianResidual(PetscTAO, PetscVec, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobianState(PetscTAO, PetscVec, PetscMat, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobianDesign(PetscTAO, PetscVec, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobianEquality(PetscTAO, PetscVec, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoJacobianInequality(PetscTAO, PetscVec, PetscMat, PetscMat, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoUpdateFunction(PetscTAO, PetscInt, void*) except PETSC_ERR_PYTHON

    ctypedef enum PetscTAOBNCGType "TaoBNCGType":
        TAO_BNCG_GD
        TAO_BNCG_PCGD
        TAO_BNCG_HS
        TAO_BNCG_FR
        TAO_BNCG_PRP
        TAO_BNCG_PRP_PLUS
        TAO_BNCG_DY
        TAO_BNCG_HZ
        TAO_BNCG_DK
        TAO_BNCG_KD
        TAO_BNCG_SSML_BFGS
        TAO_BNCG_SSML_DFP
        TAO_BNCG_SSML_BRDN

    ctypedef enum PetscTAOALMMType "TaoALMMType":
        TAO_ALMM_CLASSIC
        TAO_ALMM_PHR

    PetscErrorCode TaoMonitor(PetscTAO, PetscInt, PetscReal, PetscReal, PetscReal, PetscReal)
    PetscErrorCode TaoView(PetscTAO, PetscViewer)
    PetscErrorCode TaoDestroy(PetscTAO*)
    PetscErrorCode TaoCreate(MPI_Comm, PetscTAO*)
    PetscErrorCode TaoSetOptionsPrefix(PetscTAO, char[])
    PetscErrorCode TaoAppendOptionsPrefix(PetscTAO, char[])
    PetscErrorCode TaoGetOptionsPrefix(PetscTAO, char*[])
    PetscErrorCode TaoSetFromOptions(PetscTAO)
    PetscErrorCode TaoSetType(PetscTAO, PetscTAOType)
    PetscErrorCode TaoGetType(PetscTAO, PetscTAOType*)

    PetscErrorCode TaoSetUp(PetscTAO)
    PetscErrorCode TaoSolve(PetscTAO)

    PetscErrorCode TaoSetTolerances(PetscTAO, PetscReal, PetscReal, PetscReal)
    PetscErrorCode TaoParametersInitialize(PetscTAO)
    PetscErrorCode TaoGetTolerances(PetscTAO, PetscReal*, PetscReal*, PetscReal*)
    PetscErrorCode TaoSetConstraintTolerances(PetscTAO, PetscReal, PetscReal)
    PetscErrorCode TaoGetConstraintTolerances(PetscTAO, PetscReal*, PetscReal*)

    PetscErrorCode TaoSetFunctionLowerBound(PetscTAO, PetscReal)
    PetscErrorCode TaoSetMaximumIterations(PetscTAO, PetscInt)
    PetscErrorCode TaoGetMaximumIterations(PetscTAO, PetscInt*)
    PetscErrorCode TaoSetMaximumFunctionEvaluations(PetscTAO, PetscInt)
    PetscErrorCode TaoGetMaximumFunctionEvaluations(PetscTAO, PetscInt*)
    PetscErrorCode TaoSetIterationNumber(PetscTAO, PetscInt)
    PetscErrorCode TaoGetIterationNumber(PetscTAO, PetscInt*)

    PetscErrorCode TaoSetTrustRegionTolerance(PetscTAO, PetscReal)
    PetscErrorCode TaoGetInitialTrustRegionRadius(PetscTAO, PetscReal*)
    PetscErrorCode TaoGetTrustRegionRadius(PetscTAO, PetscReal*)
    PetscErrorCode TaoSetTrustRegionRadius(PetscTAO, PetscReal)

    PetscErrorCode TaoDefaultConvergenceTest(PetscTAO, void*) except PETSC_ERR_PYTHON
    PetscErrorCode TaoSetConvergenceTest(PetscTAO, PetscTaoConvergenceTest*, void*)
    PetscErrorCode TaoSetConvergedReason(PetscTAO, PetscTAOConvergedReason)
    PetscErrorCode TaoGetConvergedReason(PetscTAO, PetscTAOConvergedReason*)
    PetscErrorCode TaoLogConvergenceHistory(PetscTAO, PetscReal, PetscReal, PetscReal, PetscInt)
    PetscErrorCode TaoGetSolutionStatus(PetscTAO, PetscInt*,
                                        PetscReal*, PetscReal*,
                                        PetscReal*, PetscReal*,
                                        PetscTAOConvergedReason*)

    PetscErrorCode TaoMonitorSet(PetscTAO, PetscTaoMonitor, void*, PetscTaoMonitorDestroy)
    PetscErrorCode TaoMonitorCancel(PetscTAO)

    PetscErrorCode TaoComputeObjective(PetscTAO, PetscVec, PetscReal*)
    PetscErrorCode TaoComputeResidual(PetscTAO, PetscVec, PetscVec)
    PetscErrorCode TaoComputeGradient(PetscTAO, PetscVec, PetscVec)
    PetscErrorCode TaoComputeObjectiveAndGradient(PetscTAO, PetscVec, PetscReal*, PetscVec)
    PetscErrorCode TaoComputeConstraints(PetscTAO, PetscVec, PetscVec)
    PetscErrorCode TaoComputeDualVariables(PetscTAO, PetscVec, PetscVec)
    PetscErrorCode TaoComputeVariableBounds(PetscTAO)
    PetscErrorCode TaoComputeHessian(PetscTAO, PetscVec, PetscMat, PetscMat)
    PetscErrorCode TaoComputeJacobian(PetscTAO, PetscVec, PetscMat, PetscMat)

    PetscErrorCode TaoSetSolution(PetscTAO, PetscVec)
    PetscErrorCode TaoSetConstraintsVec(PetscTAO, PetscVec)
    PetscErrorCode TaoSetVariableBounds(PetscTAO, PetscVec, PetscVec)

    PetscErrorCode TaoGetSolution(PetscTAO, PetscVec*)
    PetscErrorCode TaoSetGradientNorm(PetscTAO, PetscMat)
    PetscErrorCode TaoGetGradientNorm(PetscTAO, PetscMat*)
    PetscErrorCode TaoLMVMSetH0(PetscTAO, PetscMat)
    PetscErrorCode TaoLMVMGetH0(PetscTAO, PetscMat*)
    PetscErrorCode TaoLMVMGetH0KSP(PetscTAO, PetscKSP*)
    PetscErrorCode TaoBNCGGetType(PetscTAO, PetscTAOBNCGType*)
    PetscErrorCode TaoBNCGSetType(PetscTAO, PetscTAOBNCGType)
    PetscErrorCode TaoGetVariableBounds(PetscTAO, PetscVec*, PetscVec*)

    PetscErrorCode TaoSetObjective(PetscTAO, PetscTaoObjective*, void*)
    PetscErrorCode TaoSetGradient(PetscTAO, PetscVec, PetscTaoGradient*, void*)
    PetscErrorCode TaoSetObjectiveAndGradient(PetscTAO, PetscVec, PetscTaoObjGrad*, void*)
    PetscErrorCode TaoSetHessian(PetscTAO, PetscMat, PetscMat, PetscTaoHessian*, void*)
    PetscErrorCode TaoGetObjective(PetscTAO, PetscTaoObjective**, void**)
    PetscErrorCode TaoGetGradient(PetscTAO, PetscVec*, PetscTaoGradient**, void**)
    PetscErrorCode TaoGetObjectiveAndGradient(PetscTAO, PetscVec*, PetscTaoObjGrad**, void**)
    PetscErrorCode TaoGetHessian(PetscTAO, PetscMat*, PetscMat*, PetscTaoHessian**, void**)
    PetscErrorCode TaoSetResidualRoutine(PetscTAO, PetscVec, PetscTaoResidual, void*)
    PetscErrorCode TaoSetVariableBoundsRoutine(PetscTAO, PetscTaoVarBounds*, void*)
    PetscErrorCode TaoSetConstraintsRoutine(PetscTAO, PetscVec, PetscTaoConstraints*, void*)
    PetscErrorCode TaoSetJacobianRoutine(PetscTAO, PetscMat, PetscMat, PetscTaoJacobian*, void*)
    PetscErrorCode TaoSetJacobianResidualRoutine(PetscTAO, PetscMat, PetscMat, PetscTaoJacobianResidual*, void*)
    PetscErrorCode TaoSetStateDesignIS(PetscTAO, PetscIS, PetscIS)
    PetscErrorCode TaoSetJacobianStateRoutine(PetscTAO, PetscMat, PetscMat, PetscMat, PetscTaoJacobianState*, void*)
    PetscErrorCode TaoSetJacobianDesignRoutine(PetscTAO, PetscMat, PetscTaoJacobianDesign*, void*)
    PetscErrorCode TaoGetLMVMMatrix(PetscTAO, PetscMat*)
    PetscErrorCode TaoSetLMVMMatrix(PetscTAO, PetscMat)

    PetscErrorCode TaoSetEqualityConstraintsRoutine(PetscTAO, PetscVec, PetscTaoEqualityConstraints*, void*)
    PetscErrorCode TaoGetEqualityConstraintsRoutine(PetscTAO, PetscVec*, PetscTaoEqualityConstraints**, void**)
    PetscErrorCode TaoSetJacobianEqualityRoutine(PetscTAO, PetscMat, PetscMat, PetscTaoJacobianEquality*, void*)
    PetscErrorCode TaoGetJacobianEqualityRoutine(PetscTAO, PetscMat*, PetscMat*, PetscTaoJacobianEquality**, void**)
    PetscErrorCode TaoSetInequalityConstraintsRoutine(PetscTAO, PetscVec, PetscTaoInequalityConstraints*, void*)
    PetscErrorCode TaoGetInequalityConstraintsRoutine(PetscTAO, PetscVec*, PetscTaoInequalityConstraints**, void**)
    PetscErrorCode TaoSetJacobianInequalityRoutine(PetscTAO, PetscMat, PetscMat, PetscTaoJacobianInequality*, void*)
    PetscErrorCode TaoGetJacobianInequalityRoutine(PetscTAO, PetscMat*, PetscMat*, PetscTaoJacobianInequality**, void**)
    PetscErrorCode TaoSetUpdate(PetscTAO, PetscTaoUpdateFunction*, void*)

    PetscErrorCode TaoSetInitialTrustRegionRadius(PetscTAO, PetscReal)

    PetscErrorCode TaoGetKSP(PetscTAO, PetscKSP*)
    PetscErrorCode TaoGetLineSearch(PetscTAO, PetscTAOLineSearch*)

    PetscErrorCode TaoALMMGetSubsolver(PetscTAO, PetscTAO*)
    PetscErrorCode TaoALMMSetSubsolver(PetscTAO, PetscTAO)
    PetscErrorCode TaoALMMGetType(PetscTAO, PetscTAOALMMType*)
    PetscErrorCode TaoALMMSetType(PetscTAO, PetscTAOALMMType)

    PetscErrorCode TaoBRGNGetSubsolver(PetscTAO, PetscTAO*)
    PetscErrorCode TaoBRGNSetRegularizerObjectiveAndGradientRoutine(PetscTAO, PetscTaoRegularizerObjGrad*, void*)
    PetscErrorCode TaoBRGNSetRegularizerHessianRoutine(PetscTAO, PetscMat, PetscTaoRegularizerHessian*, void*)
    PetscErrorCode TaoBRGNSetRegularizerWeight(PetscTAO, PetscReal)
    PetscErrorCode TaoBRGNSetL1SmoothEpsilon(PetscTAO, PetscReal)
    PetscErrorCode TaoBRGNSetDictionaryMatrix(PetscTAO, PetscMat)
    PetscErrorCode TaoBRGNGetDampingVector(PetscTAO, PetscVec*)

    PetscErrorCode TaoPythonSetType(PetscTAO, char[])
    PetscErrorCode TaoPythonGetType(PetscTAO, char*[])

    ctypedef const char* PetscTAOLineSearchType "TaoLineSearchType"
    PetscTAOLineSearchType TAOLINESEARCHUNIT
    PetscTAOLineSearchType TAOLINESEARCHARMIJO
    PetscTAOLineSearchType TAOLINESEARCHOWARMIJO
    PetscTAOLineSearchType TAOLINESEARCHGPCG
    PetscTAOLineSearchType TAOLINESEARCHMT
    PetscTAOLineSearchType TAOLINESEARCHIPM

    ctypedef enum PetscTAOLineSearchConvergedReason "TaoLineSearchConvergedReason":
        # failed
        TAOLINESEARCH_FAILED_INFORNAN
        TAOLINESEARCH_FAILED_BADPARAMETER
        TAOLINESEARCH_FAILED_ASCENT
        # continue
        TAOLINESEARCH_CONTINUE_ITERATING
        # success
        TAOLINESEARCH_SUCCESS
        TAOLINESEARCH_SUCCESS_USER
        # halted
        TAOLINESEARCH_HALTED_OTHER
        TAOLINESEARCH_HALTED_MAXFCN
        TAOLINESEARCH_HALTED_UPPERBOUND
        TAOLINESEARCH_HALTED_LOWERBOUND
        TAOLINESEARCH_HALTED_RTOL
        TAOLINESEARCH_HALTED_USER

    ctypedef PetscErrorCode PetscTaoLineSearchObjective(PetscTAOLineSearch, PetscVec, PetscReal*, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoLineSearchGradient(PetscTAOLineSearch, PetscVec, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoLineSearchObjGrad(PetscTAOLineSearch, PetscVec, PetscReal*, PetscVec, void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode PetscTaoLineSearchObjGTS(PetscTaoLineSearch, PetscVec, PetscVec, PetscReal*, PetscReal*, void*) except PETSC_ERR_PYTHON

    PetscErrorCode TaoLineSearchCreate(MPI_Comm, PetscTAOLineSearch*)
    PetscErrorCode TaoLineSearchDestroy(PetscTAOLineSearch*)
    PetscErrorCode TaoLineSearchView(PetscTAOLineSearch, PetscViewer)
    PetscErrorCode TaoLineSearchSetType(PetscTAOLineSearch, PetscTAOLineSearchType)
    PetscErrorCode TaoLineSearchGetType(PetscTAOLineSearch, PetscTAOLineSearchType*)
    PetscErrorCode TaoLineSearchSetOptionsPrefix(PetscTAOLineSearch, char[])
    PetscErrorCode TaoLineSearchGetOptionsPrefix(PetscTAOLineSearch, char*[])
    PetscErrorCode TaoLineSearchSetFromOptions(PetscTAOLineSearch)
    PetscErrorCode TaoLineSearchSetUp(PetscTAOLineSearch)
    PetscErrorCode TaoLineSearchUseTaoRoutines(PetscTAOLineSearch, PetscTAO)
    PetscErrorCode TaoLineSearchSetObjectiveRoutine(PetscTAOLineSearch, PetscTaoLineSearchObjective, void*)
    PetscErrorCode TaoLineSearchSetGradientRoutine(PetscTAOLineSearch, PetscTaoLineSearchGradient, void*)
    PetscErrorCode TaoLineSearchSetObjectiveAndGradientRoutine(PetscTAOLineSearch, PetscTaoLineSearchObjGrad, void*)
    PetscErrorCode TaoLineSearchApply(PetscTAOLineSearch, PetscVec, PetscReal*, PetscVec, PetscVec, PetscReal*, PetscTAOLineSearchConvergedReason*)
    PetscErrorCode TaoLineSearchSetInitialStepLength(PetscTAOLineSearch, PetscReal)

# --------------------------------------------------------------------

cdef inline TAO ref_TAO(PetscTAO tao):
    cdef TAO ob = <TAO> TAO()
    ob.tao = tao
    CHKERR(PetscINCREF(ob.obj))
    return ob

# --------------------------------------------------------------------

cdef PetscErrorCode TAO_Objective(PetscTAO _tao,
                                  PetscVec _x, PetscReal *_f,
                                  void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    context = tao.get_attr("__objective__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (objective, args, kargs) = context
    retv = objective(tao, x, *args, **kargs)
    _f[0] = asReal(retv)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Residual(PetscTAO _tao,
                                 PetscVec _x, PetscVec _r,
                                 void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec r   = ref_Vec(_r)
    context = tao.get_attr("__residual__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (residual, args, kargs) = context
    residual(tao, x, r, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Gradient(PetscTAO _tao,
                                 PetscVec _x, PetscVec _g,
                                 void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec g   = ref_Vec(_g)
    context = tao.get_attr("__gradient__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (gradient, args, kargs) = context
    gradient(tao, x, g, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_ObjGrad(PetscTAO _tao,
                                PetscVec _x, PetscReal *_f, PetscVec _g,
                                void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec g   = ref_Vec(_g)
    context = tao.get_attr("__objgrad__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (objgrad, args, kargs) = context
    retv = objgrad(tao, x, g, *args, **kargs)
    _f[0] = asReal(retv)
    return PETSC_SUCCESS


cdef PetscErrorCode TAO_BRGNRegObjGrad(PetscTAO _tao,
                                       PetscVec _x, PetscReal *_f, PetscVec _g,
                                       void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec g   = ref_Vec(_g)
    context = tao.get_attr("__brgnregobjgrad__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (objgrad, args, kargs) = context
    retv = objgrad(tao, x, g, *args, **kargs)
    _f[0] = asReal(retv)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Constraints(PetscTAO _tao,
                                    PetscVec _x, PetscVec _r,
                                    void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec r   = ref_Vec(_r)
    context = tao.get_attr("__constraints__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (constraints, args, kargs) = context
    constraints(tao, x, r, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_VarBounds(PetscTAO _tao,
                                  PetscVec _xl, PetscVec _xu,
                                  void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAO tao = ref_TAO(_tao)
    cdef Vec xl  = ref_Vec(_xl)
    cdef Vec xu  = ref_Vec(_xu)
    context = tao.get_attr("__varbounds__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (varbounds, args, kargs) = context
    varbounds(tao, xl, xu, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Hessian(PetscTAO _tao,
                                PetscVec _x,
                                PetscMat _H,
                                PetscMat _P,
                                void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat H   = ref_Mat(_H)
    cdef Mat P   = ref_Mat(_P)
    context = tao.get_attr("__hessian__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (hessian, args, kargs) = context
    hessian(tao, x, H, P, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_BRGNRegHessian(PetscTAO _tao,
                                       PetscVec  _x,
                                       PetscMat  _H,
                                       void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat H   = ref_Mat(_H)
    context = tao.get_attr("__brgnreghessian__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (hessian, args, kargs) = context
    hessian(tao, x, H, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Jacobian(PetscTAO _tao,
                                 PetscVec  _x,
                                 PetscMat  _J,
                                 PetscMat  _P,
                                 void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    cdef Mat P   = ref_Mat(_P)
    context = tao.get_attr("__jacobian__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, P, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_JacobianResidual(PetscTAO _tao,
                                         PetscVec  _x,
                                         PetscMat  _J,
                                         PetscMat  _P,
                                         void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    cdef Mat P   = ref_Mat(_P)
    context = tao.get_attr("__jacobian_residual__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, P, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_JacobianState(PetscTAO _tao,
                                      PetscVec  _x,
                                      PetscMat  _J,
                                      PetscMat  _Jp,
                                      PetscMat  _Ji,
                                      void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    cdef Mat Jp  = ref_Mat(_Jp)
    cdef Mat Ji  = ref_Mat(_Ji)
    context = tao.get_attr("__jacobian_state__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, Jp, Ji, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_JacobianDesign(PetscTAO _tao,
                                       PetscVec  _x,
                                       PetscMat  _J,
                                       void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    context = tao.get_attr("__jacobian_design__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_EqualityConstraints(PetscTAO _tao,
                                            PetscVec  _x,
                                            PetscVec  _c,
                                            void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec c   = ref_Vec(_c)
    context = tao.get_attr("__equality_constraints__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (f, args, kargs) = context
    f(tao, x, c, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_InequalityConstraints(PetscTAO _tao,
                                              PetscVec  _x,
                                              PetscVec  _c,
                                              void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec c   = ref_Vec(_c)
    context = tao.get_attr("__inequality_constraints__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (f, args, kargs) = context
    f(tao, x, c, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_JacobianEquality(PetscTAO _tao,
                                         PetscVec  _x,
                                         PetscMat  _J,
                                         PetscMat  _P,
                                         void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    cdef Mat P   = ref_Mat(_P)
    context = tao.get_attr("__jacobian_equality__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, P, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_JacobianInequality(PetscTAO _tao,
                                           PetscVec  _x,
                                           PetscMat  _J,
                                           PetscMat  _P,
                                           void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef Vec x   = ref_Vec(_x)
    cdef Mat J   = ref_Mat(_J)
    cdef Mat P   = ref_Mat(_P)
    context = tao.get_attr("__jacobian_inequality__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple # sanity check
    (jacobian, args, kargs) = context
    jacobian(tao, x, J, P, *args, **kargs)
    return PETSC_SUCCESS

# ctx is unused
cdef PetscErrorCode TAO_Update(
    PetscTAO _tao,
    PetscInt its,
    void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    (update, args, kargs) = tao.get_attr('__update__')
    update(tao, toInt(its), *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Converged(PetscTAO _tao,
                                  void* ctx) except PETSC_ERR_PYTHON with gil:
    # call first the default convergence test
    CHKERR(TaoDefaultConvergenceTest(_tao, NULL))
    # call next the user-provided convergence test
    cdef TAO tao = ref_TAO(_tao)
    (converged, args, kargs) = tao.get_attr('__converged__')
    reason = converged(tao, *args, **kargs)
    if reason is None:  return PETSC_SUCCESS
    # handle value of convergence reason
    cdef PetscTAOConvergedReason creason = TAO_CONTINUE_ITERATING
    if reason is False or reason == -1:
        creason = TAO_DIVERGED_USER
    elif reason is True or reason == 1:
        creason = TAO_CONVERGED_USER
    else:
        creason = reason
        assert creason >= TAO_DIVERGED_USER
        assert creason <= TAO_CONVERGED_USER
    CHKERR(TaoSetConvergedReason(_tao, creason))
    return PETSC_SUCCESS

cdef PetscErrorCode TAO_Monitor(PetscTAO _tao,
                                void* ctx) except PETSC_ERR_PYTHON with gil:
    cdef TAO tao = ref_TAO(_tao)
    cdef object monitorlist = tao.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    for (monitor, args, kargs) in monitorlist:
        monitor(tao, *args, **kargs)
    return PETSC_SUCCESS

# --------------------------------------------------------------------

cdef inline TAOLineSearch ref_TAOLS(PetscTAOLineSearch taols):
    cdef TAOLineSearch ob = <TAOLineSearch> TAOLineSearch()
    ob.taols = taols
    CHKERR(PetscINCREF(ob.obj))
    return ob

# --------------------------------------------------------------------

cdef PetscErrorCode TAOLS_Objective(PetscTAOLineSearch _ls,
                                    PetscVec _x,
                                    PetscReal *_f,
                                    void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAOLineSearch ls = ref_TAOLS(_ls)
    cdef Vec x   = ref_Vec(_x)
    (objective, args, kargs) = ls.get_attr("__objective__")
    retv = objective(ls, x, *args, **kargs)
    _f[0] = asReal(retv)
    return PETSC_SUCCESS

cdef PetscErrorCode TAOLS_Gradient(PetscTAOLineSearch _ls,
                                   PetscVec _x,
                                   PetscVec _g,
                                   void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAOLineSearch ls = ref_TAOLS(_ls)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec g   = ref_Vec(_g)
    (gradient, args, kargs) = ls.get_attr("__gradient__")
    gradient(ls, x, g, *args, **kargs)
    return PETSC_SUCCESS


cdef PetscErrorCode TAOLS_ObjGrad(PetscTAOLineSearch _ls,
                                  PetscVec _x,
                                  PetscReal *_f,
                                  PetscVec _g,
                                  void *ctx) except PETSC_ERR_PYTHON with gil:

    cdef TAOLineSearch ls = ref_TAOLS(_ls)
    cdef Vec x   = ref_Vec(_x)
    cdef Vec g   = ref_Vec(_g)
    (objgrad, args, kargs) = ls.get_attr("__objgrad__")
    retv = objgrad(ls, x, g, *args, **kargs)
    _f[0] = asReal(retv)
    return PETSC_SUCCESS



# --------------------------------------------------------------------

cdef class Quad(Object):
    """Quadrature rule for integration."""
    def __cinit__(self):
        self.obj = <PetscObject*> &self.quad
        self.quad = NULL

    def view(self, Viewer viewer=None) -> None:
        """View a `Quad` object.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the graph.

        See Also
        --------
        petsc.PetscQuadratureView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscQuadratureView(self.quad, vwr))

    def create(self, comm: Comm | None = None) -> Self:
        """Create a `Quad` object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscQuadratureCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscQuadrature newquad = NULL
        CHKERR(PetscQuadratureCreate(ccomm, &newquad))
        CHKERR(PetscCLEAR(self.obj)); self.quad = newquad
        return self

    def duplicate(self) -> Quad:
        """Create a deep copy of the `Quad` object.

        Collective.

        See Also
        --------
        petsc.PetscQuadratureDuplicate

        """
        cdef Quad newquad = Quad()
        CHKERR(PetscQuadratureDuplicate(self.quad, &newquad.quad))
        return newquad

    def destroy(self) -> Self:
        """Destroy the `Quad` object.

        Collective.

        See Also
        --------
        petsc.PetscQuadratureDestroy

        """
        CHKERR(PetscQuadratureDestroy(&self.quad))
        return self

    def getData(self) -> tuple[ArrayReal, ArrayReal]:
        """Return the data defining the `Quad`.

        Not collective.

        Returns
        -------
        points : ArrayReal
            The coordinates of the quadrature points.
        weights : ArrayReal
            The quadrature weights.

        See Also
        --------
        petsc.PetscQuadratureGetData

        """
        cdef PetscInt cdim = 0
        cdef PetscInt cnc = 0
        cdef PetscInt cnpoints = 0
        cdef const PetscReal *cpoints = NULL
        cdef const PetscReal *cweights = NULL
        CHKERR(PetscQuadratureGetData(self.quad, &cdim, &cnc, &cnpoints, &cpoints, &cweights))
        return array_r(cnpoints*cdim, cpoints), array_r(cnpoints*cnc, cweights)

    # FIXME:
    # def setData(???)

    def getNumComponents(self) -> int:
        """Return the number of components for functions to be integrated.

        Not collective.

        See Also
        --------
        setNumComponents, petsc.PetscQuadratureGetNumComponents

        """
        cdef PetscInt cnc = 0
        CHKERR(PetscQuadratureGetNumComponents(self.quad, &cnc))
        return toInt(cnc)

    def setNumComponents(self, nc: int) -> None:
        """Return the number of components for functions to be integrated.

        Not collective.

        Parameters
        ----------
        nc
            The number of components.

        See Also
        --------
        getNumComponents, petsc.PetscQuadratureSetNumComponents

        """
        cdef PetscInt cnc = asInt(nc)
        CHKERR(PetscQuadratureSetNumComponents(self.quad, cnc))

    def getOrder(self) -> int:
        """Return the order of the method in the `Quad`.

        Not collective.

        See Also
        --------
        setOrder, petsc.PetscQuadratureGetOrder

        """
        cdef PetscInt corder = 0
        CHKERR(PetscQuadratureGetOrder(self.quad, &corder))
        return toInt(corder)

    def setOrder(self, order: int) -> None:
        """Set the order of the method in the `Quad`.

        Not collective.

        Parameters
        ----------
        order
            The order of the quadrature, i.e. the highest degree polynomial
            that is exactly integrated.

        See Also
        --------
        getOrder, petsc.PetscQuadratureSetOrder

        """
        cdef PetscInt corder = asInt(order)
        CHKERR(PetscQuadratureSetOrder(self.quad, corder))


# --------------------------------------------------------------------



# --------------------------------------------------------------------

from petsc4py import PETSc
import unittest
from test_ksp import BaseTestKSP

# --------------------------------------------------------------------


class MyKSP:
    def __init__(self):
        pass

    def create(self, ksp):
        self.work = []

    def destroy(self, ksp):
        for v in self.work:
            v.destroy()

    def setUp(self, ksp):
        self.work[:] = ksp.getWorkVecs(right=2, left=None)

    def reset(self, ksp):
        for v in self.work:
            v.destroy()
        del self.work[:]

    def loop(self, ksp, r):
        its = ksp.getIterationNumber()
        rnorm = r.norm()
        ksp.setResidualNorm(rnorm)
        ksp.logConvergenceHistory(rnorm)
        ksp.monitor(its, rnorm)
        reason = ksp.callConvergenceTest(its, rnorm)
        if not reason:
            ksp.setIterationNumber(its + 1)
        else:
            ksp.setConvergedReason(reason)
        return reason


class MyRichardson(MyKSP):
    def solve(self, ksp, b, x):
        A, B = ksp.getOperators()
        P = ksp.getPC()
        r, z = self.work
        #
        A.mult(x, r)
        r.aypx(-1, b)
        P.apply(r, z)
        x.axpy(1, z)
        while not self.loop(ksp, z):
            A.mult(x, r)
            r.aypx(-1, b)
            P.apply(r, z)
            x.axpy(1, z)


class MyCG(MyKSP):
    def setUp(self, ksp):
        super().setUp(ksp)
        d = self.work[0].duplicate()
        q = d.duplicate()
        self.work += [d, q]

    def solve(self, ksp, b, x):
        A, B = ksp.getOperators()
        # P = ksp.getPC()
        r, z, d, q = self.work
        #
        A.mult(x, r)
        r.aypx(-1, b)
        r.copy(d)
        delta_0 = r.dot(r)
        delta = delta_0
        while not self.loop(ksp, r):
            A.mult(d, q)
            alpha = delta / d.dot(q)
            x.axpy(+alpha, d)
            r.axpy(-alpha, q)
            delta_old = delta
            delta = r.dot(r)
            beta = delta / delta_old
            d.aypx(beta, r)


# --------------------------------------------------------------------


class BaseTestKSPPYTHON(BaseTestKSP):
    KSP_TYPE = PETSc.KSP.Type.PYTHON
    ContextClass = None

    def setUp(self):
        super().setUp()
        ctx = self.ContextClass()
        self.ksp.setPythonContext(ctx)

    def testGetType(self):
        ctx = self.ksp.getPythonContext()
        pytype = f'{ctx.__module__}.{type(ctx).__name__}'
        self.assertTrue(self.ksp.getPythonType() == pytype)

    def tearDown(self):
        self.ksp.destroy()
        PETSc.garbage_cleanup()


class TestKSPPYTHON_RICH(BaseTestKSPPYTHON, unittest.TestCase):
    PC_TYPE = PETSc.PC.Type.JACOBI
    ContextClass = MyRichardson


class TestKSPPYTHON_CG(BaseTestKSPPYTHON, unittest.TestCase):
    PC_TYPE = PETSc.PC.Type.NONE
    ContextClass = MyCG


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()



# --------------------------------------------------------------------

cdef extern from * nogil:

    struct _n_PetscLayout
    ctypedef _n_PetscLayout* PetscLayout
    PetscErrorCode PetscLayoutSetLocalSize(PetscLayout, PetscInt)
    PetscErrorCode PetscLayoutSetSize(PetscLayout, PetscInt)
    PetscErrorCode PetscLayoutGetBlockSize(PetscLayout, PetscInt*)
    PetscErrorCode PetscLayoutSetBlockSize(PetscLayout, PetscInt)
    PetscErrorCode PetscLayoutSetUp(PetscLayout)

# --------------------------------------------------------------------



import unittest
from petsc4py import PETSc
from sys import getrefcount
import numpy


# --------------------------------------------------------------------
class Objective:
    def __call__(self, tao, x):
        return (x[0] - 1.0) ** 2 + (x[1] - 2.0) ** 2


class Gradient:
    def __call__(self, tao, x, g):
        g[0] = 2.0 * (x[0] - 1.0)
        g[1] = 2.0 * (x[1] - 2.0)
        g.assemble()


class MyTao:
    def __init__(self):
        self.log = {}

    def _log(self, method):
        self.log.setdefault(method, 0)
        self.log[method] += 1

    def create(self, tao):
        self._log('create')
        self.testvec = PETSc.Vec()

    def destroy(self, tao):
        self._log('destroy')
        self.testvec.destroy()

    def setFromOptions(self, tao):
        self._log('setFromOptions')

    def setUp(self, tao):
        self._log('setUp')
        self.testvec = tao.getSolution().duplicate()

    def solve(self, tao):
        self._log('solve')

    def step(self, tao, x, g, s):
        self._log('step')
        tao.computeGradient(x, g)
        g.copy(s)
        s.scale(-1.0)

    def preStep(self, tao):
        self._log('preStep')

    def postStep(self, tao):
        self._log('postStep')

    def monitor(self, tao):
        self._log('monitor')


class TestTaoPython(unittest.TestCase):
    def setUp(self):
        self.tao = PETSc.TAO()
        self.tao.createPython(MyTao(), comm=PETSc.COMM_SELF)
        self.assertEqual(getrefcount(self._getCtx()), 2)
        self.assertEqual(self._getCtx().log['create'], 1)
        self.nsolve = 0

    def tearDown(self):
        self.assertEqual(getrefcount(self._getCtx()), 2)
        self.assertTrue('destroy' not in self._getCtx().log)
        ctx = self._getCtx()
        self.tao.destroy()
        self.tao = None
        PETSc.garbage_cleanup()
        self.assertEqual(ctx.log['destroy'], 1)

    def testGetType(self):
        ctx = self.tao.getPythonContext()
        pytype = f'{ctx.__module__}.{type(ctx).__name__}'
        self.assertTrue(self.tao.getPythonType() == pytype)

    def testSolve(self):
        tao = self.tao
        ctx = tao.getPythonContext()
        x = PETSc.Vec().create(tao.getComm())
        x.setType('standard')
        x.setSizes(2)
        y1 = x.duplicate()
        y2 = x.duplicate()
        tao.setObjective(Objective())
        tao.setGradient(Gradient(), None)
        tao.setMonitor(ctx.monitor)
        tao.setFromOptions()
        tao.setMaximumIterations(3)

        def _update(tao, it, cnt):
            cnt += 1

        cnt_up = numpy.array(0)
        tao.setUpdate(_update, (cnt_up,))
        tao.setSolution(x)

        # Call the solve method of MyTAO
        x.set(0.5)
        tao.solve()
        n = tao.getIterationNumber()
        self.assertTrue(n == 0)

        # Call the default solve method and use step of MyTAO
        ctx.solve = None
        x.set(0.5)
        tao.solve()
        n = tao.getIterationNumber()
        self.assertGreater(tao.getConvergedReason(), 0)
        self.assertTrue(n in [2, 3])
        self.assertAlmostEqual(x[0], 1.0)
        self.assertAlmostEqual(x[1], 2.0)

        # Call the default solve method with the default step method
        ctx.step = None
        x.set(0.5)
        tao.solve()
        n = tao.getIterationNumber()
        self.assertGreater(tao.getConvergedReason(), 0)
        self.assertTrue(n in [2, 3])
        self.assertAlmostEqual(x[0], 1.0)
        self.assertAlmostEqual(x[1], 2.0)

        self.assertTrue(y1.equal(y2))
        self.assertTrue(ctx.log['monitor'] == 2 * (n + 1))
        self.assertTrue(ctx.log['preStep'] == 2 * n)
        self.assertTrue(ctx.log['postStep'] == 2 * n)
        self.assertTrue(ctx.log['solve'] == 1)
        self.assertTrue(ctx.log['setUp'] == 1)
        self.assertTrue(ctx.log['setFromOptions'] == 1)
        self.assertTrue(ctx.log['step'] == n)
        self.assertEqual(cnt_up, 2 * n)
        tao.cancelMonitor()

    def _getCtx(self):
        return self.tao.getPythonContext()


class MyGradientDescent:
    def __init__(self):
        self._ls = None

    def create(self, tao):
        self._ls = PETSc.TAOLineSearch().create(comm=PETSc.COMM_SELF)
        self._ls.useTAORoutine(tao)
        self._ls.setType(PETSc.TAOLineSearch.Type.UNIT)
        self._ls.setInitialStepLength(0.2)

    def destroy(self, tao):
        self._ls.destroy()

    def setUp(self, tao):
        pass

    def solve(self, tao):
        x = tao.getSolution()
        gradient = tao.getGradient()[0]
        search_direction = gradient.copy()
        for it in range(tao.getMaximumIterations()):
            tao.setIterationNumber(it)

            # search_direction = -gradient
            tao.computeGradient(x, gradient)
            gradient.copy(search_direction)
            search_direction.scale(-1)

            # x = x + .2 search_direction
            f, s, reason = self._ls.apply(x, gradient, search_direction)

            tao.monitor(f=f, res=gradient.norm())

            if reason < 0:
                raise RuntimeError('LS failed.')

            if tao.checkConverged() > 0:
                break

    def step(self, tao, x, g, s):
        raise RuntimeError('Should only be called by builtin solve.')

    def preStep(self, tao):
        raise RuntimeError('Should only be called by builtin solve.')

    def postStep(self, tao):
        raise RuntimeError('Should only be called by builtin solve.')


class TestTaoPythonOptimiser(unittest.TestCase):
    def setUp(self):
        self.tao = PETSc.TAO()
        self.tao.createPython(MyGradientDescent(), comm=PETSc.COMM_SELF)

    def tearDown(self):
        self.tao.destroy()
        self.tao = None

    def testSolve(self):
        tao = self.tao

        opts = PETSc.Options('test_tao_python_optimiser_')
        opts['tao_max_it'] = 100
        opts['tao_gatol'] = 1e-6

        tao.setOptionsPrefix('test_tao_python_optimiser_')
        tao.setFromOptions()

        x = PETSc.Vec().createSeq(2, comm=tao.getComm())
        x.set(0.5)

        tao.setSolution(x)
        tao.setObjective(Objective())
        tao.setGradient(Gradient(), x.copy())

        tao.solve()

        self.assertEqual(tao.getMaximumIterations(), 100)
        self.assertAlmostEqual(tao.getTolerances()[0], 1e-6)
        self.assertGreater(tao.getIterationNumber(), 0)
        self.assertGreater(tao.getConvergedReason(), 0)
        self.assertAlmostEqual(x[0], 1.0, places=5)
        self.assertAlmostEqual(x[1], 2.0, places=5)
        self.assertGreater(tao.getObjectiveValue(), 0)
        self.assertAlmostEqual(tao.getObjectiveValue(), 0, places=5)


# --------------------------------------------------------------------

if numpy.iscomplexobj(PETSc.ScalarType()):
    del TestTaoPython
    del TestTaoPythonOptimiser

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef enum PetscOffloadMask:
        PETSC_OFFLOAD_UNALLOCATED
        PETSC_OFFLOAD_CPU
        PETSC_OFFLOAD_GPU
        PETSC_OFFLOAD_BOTH
        PETSC_OFFLOAD_KOKKOS

    ctypedef enum PetscMemType:
        PETSC_MEMTYPE_HOST
        PETSC_MEMTYPE_CUDA
        PETSC_MEMTYPE_HIP
        PETSC_MEMTYPE_SYCL

    ctypedef enum PetscDeviceType:
        PETSC_DEVICE_HOST
        PETSC_DEVICE_CUDA
        PETSC_DEVICE_HIP
        PETSC_DEVICE_SYCL

    ctypedef enum PetscStreamType:
        PETSC_STREAM_DEFAULT
        PETSC_STREAM_NONBLOCKING
        PETSC_STREAM_DEFAULT_WITH_BARRIER
        PETSC_STREAM_NONBLOCKING_WITH_BARRIER

    ctypedef enum PetscDeviceContextJoinMode:
        PETSC_DEVICE_CONTEXT_JOIN_DESTROY
        PETSC_DEVICE_CONTEXT_JOIN_SYNC
        PETSC_DEVICE_CONTEXT_JOIN_NO_SYNC

    PetscErrorCode PetscDeviceCreate(PetscDeviceType, PetscInt, PetscDevice *)
    PetscErrorCode PetscDeviceDestroy(PetscDevice *)
    PetscErrorCode PetscDeviceConfigure(PetscDevice)
    PetscErrorCode PetscDeviceView(PetscDevice, PetscViewer)
    PetscErrorCode PetscDeviceGetType(PetscDevice, PetscDeviceType *)
    PetscErrorCode PetscDeviceGetDeviceId(PetscDevice, PetscInt *)
    PetscDeviceType PETSC_DEVICE_DEFAULT()
    PetscErrorCode PetscDeviceSetDefaultDeviceType(PetscDeviceType)
    PetscErrorCode PetscDeviceInitialize(PetscDeviceType)
    PetscBool PetscDeviceInitialized(PetscDeviceType)

    PetscErrorCode PetscDeviceContextCreate(PetscDeviceContext *)
    PetscErrorCode PetscDeviceContextDestroy(PetscDeviceContext *)
    PetscErrorCode PetscDeviceContextSetStreamType(PetscDeviceContext, PetscStreamType)
    PetscErrorCode PetscDeviceContextGetStreamType(PetscDeviceContext, PetscStreamType *)
    PetscErrorCode PetscDeviceContextSetDevice(PetscDeviceContext, PetscDevice)
    PetscErrorCode PetscDeviceContextGetDevice(PetscDeviceContext, PetscDevice *)
    PetscErrorCode PetscDeviceContextGetDeviceType(PetscDeviceContext, PetscDeviceType *)
    PetscErrorCode PetscDeviceContextSetUp(PetscDeviceContext)
    PetscErrorCode PetscDeviceContextDuplicate(PetscDeviceContext, PetscDeviceContext *)
    PetscErrorCode PetscDeviceContextQueryIdle(PetscDeviceContext, PetscBool *)
    PetscErrorCode PetscDeviceContextWaitForContext(PetscDeviceContext, PetscDeviceContext)
    PetscErrorCode PetscDeviceContextForkWithStreamType(PetscDeviceContext, PetscStreamType, PetscInt, PetscDeviceContext **)
    PetscErrorCode PetscDeviceContextFork(PetscDeviceContext, PetscInt, PetscDeviceContext **)
    PetscErrorCode PetscDeviceContextJoin(PetscDeviceContext, PetscInt, PetscDeviceContextJoinMode, PetscDeviceContext **)
    PetscErrorCode PetscDeviceContextSynchronize(PetscDeviceContext)
    PetscErrorCode PetscDeviceContextSetFromOptions(MPI_Comm, PetscDeviceContext)
    PetscErrorCode PetscDeviceContextView(PetscDeviceContext, PetscViewer)
    PetscErrorCode PetscDeviceContextViewFromOptions(PetscDeviceContext, PetscObject, const char name[])
    PetscErrorCode PetscDeviceContextGetCurrentContext(PetscDeviceContext *)
    PetscErrorCode PetscDeviceContextSetCurrentContext(PetscDeviceContext)

cdef extern from * nogil: # custom.h
    PetscErrorCode PetscDeviceReference(PetscDevice)

cdef inline PetscDeviceType asDeviceType(object dtype) except <PetscDeviceType>(-1):
    if isinstance(dtype, str):
        dtype = dtype.upper()
        try:
            return getattr(Device.Type, dtype)
        except AttributeError:
            raise ValueError("unknown device type: %s" % dtype)
    return dtype

cdef inline str toDeviceType(PetscDeviceType dtype):
    try:
        return Device.Type.__enum2str[dtype]
    except KeyError:
        raise NotImplementedError("unhandled PetscDeviceType %d" % <int>dtype)

cdef inline PetscStreamType asStreamType(object stype) except <PetscStreamType>(-1):
    if isinstance(stype, str):
        stype = stype.upper()
        try:
            return getattr(DeviceContext.StreamType, stype)
        except AttributeError:
            raise ValueError("unknown stream type: %s" % stype)
    return stype

cdef inline str toStreamType(PetscStreamType stype):
    try:
        return DeviceContext.StreamType.__enum2str[stype]
    except KeyError:
        raise NotImplementedError("unhandled PetscStreamType %d" % <int>stype)

cdef inline PetscDeviceContextJoinMode asJoinMode(object jmode) except <PetscDeviceContextJoinMode>(-1):
    if isinstance(jmode, str):
        jmode = jmode.upper()
        try:
            return getattr(DeviceContext.JoinMode, jmode)
        except AttributeError:
            raise ValueError("unknown join mode: %s" % jmode)
    return jmode

cdef inline str toJoinMode(PetscDeviceContextJoinMode jmode):
    try:
        return DeviceContext.JoinMode.__enum2str[jmode]
    except KeyError:
        raise NotImplementedError("unhandled PetscDeviceContextJoinMode %d" % <int>jmode)



import unittest

# --------------------------------------------------------------------


class TestStdout(unittest.TestCase):
    def testStdoutRedirect(self):
        from io import StringIO
        import sys

        prevstdout = sys.stdout
        prevstderr = sys.stderr
        sys.stdout = StringIO()
        sys.stderr = StringIO()

        import numpy as np
        from petsc4py import PETSc

        if not (__name__ == '__main__'):
            PETSc._push_python_vfprintf()

        a = np.array([0.0, 0.0, 0.0], dtype=PETSc.ScalarType)
        a_vec = PETSc.Vec().createWithArray(a, comm=PETSc.COMM_SELF)
        a_vec.view()
        v = PETSc.Viewer.STDERR(PETSc.COMM_SELF)
        v.printfASCII('Error message')

        newstdout = sys.stdout
        newstderr = sys.stderr
        sys.stdout = prevstdout
        sys.stderr = prevstderr

        output = newstdout.getvalue()
        error = newstderr.getvalue()
        if not (__name__ == '__main__'):
            PETSc._pop_python_vfprintf()
        stdoutshouldbe = """Vec Object: 1 MPI process
  type: seq
0.
0.
0.
"""
        stderrshouldbe = 'Error message'
        if PETSc._stdout_is_stderr():
            stdoutshouldbe = stdoutshouldbe + stderrshouldbe
            stderrshouldbe = ''
        self.assertEqual(output, stdoutshouldbe)
        self.assertEqual(error, stderrshouldbe)


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# --------------------------------------------------------------------

cdef extern from * nogil:

    PetscErrorCode DMCompositeCreate(MPI_Comm, PetscDM*)
    PetscErrorCode DMCompositeAddDM(PetscDM, PetscDM)
    PetscErrorCode DMCompositeGetNumberDM(PetscDM, PetscInt*)
    PetscErrorCode DMCompositeScatterArray(PetscDM, PetscVec, PetscVec*)
    PetscErrorCode DMCompositeGatherArray(PetscDM, PetscInsertMode, PetscVec, PetscVec*)
    PetscErrorCode DMCompositeGetEntriesArray(PetscDM, PetscDM*)
    PetscErrorCode DMCompositeGetAccessArray(PetscDM, PetscVec, PetscInt, const PetscInt*, PetscVec*)
    PetscErrorCode DMCompositeRestoreAccessArray(PetscDM, PetscVec, PetscInt, const PetscInt*, PetscVec*)
    PetscErrorCode DMCompositeGetGlobalISs(PetscDM, PetscIS**)
    PetscErrorCode DMCompositeGetLocalISs(PetscDM, PetscIS**)
    PetscErrorCode DMCompositeGetISLocalToGlobalMappings(PetscDM, PetscLGMap**)

cdef class _DMComposite_access:
    cdef PetscDM  dm
    cdef PetscVec gvec
    cdef PetscInt nlocs
    cdef PetscInt *locs
    cdef PetscVec *vecs
    cdef object locs_mem
    cdef object vecs_mem
    cdef object access

    def __cinit__(self, DM dm, Vec gvec, locs=None):
        self.dm = dm.dm
        CHKERR(PetscINCREF(<PetscObject*>&self.dm))
        self.gvec = gvec.vec
        CHKERR(PetscINCREF(<PetscObject*>&self.gvec))
        if locs is None:
            CHKERR(DMCompositeGetNumberDM(self.dm, &self.nlocs))
            locs = arange(0, <long>self.nlocs, 1)
        self.locs_mem = iarray_i(locs, &self.nlocs, &self.locs)
        self.vecs_mem = oarray_p(empty_p(self.nlocs), NULL, <void**>&self.vecs)
        self.access   = None

    def __dealloc__(self):
        CHKERR(DMDestroy(&self.dm))
        CHKERR(VecDestroy(&self.gvec))

    def __enter__(self):
        cdef Py_ssize_t n = self.nlocs
        CHKERR(DMCompositeGetAccessArray(self.dm, self.gvec, self.nlocs, self.locs, self.vecs))
        self.access = [ref_Vec(self.vecs[i]) for i from 0 <= i < n]
        return tuple(self.access)

    def __exit__(self, *exc):
        cdef Py_ssize_t i, n = self.nlocs
        for i from 0 <= i < n: (<Vec>self.access[i]).vec = NULL
        CHKERR(DMCompositeRestoreAccessArray(self.dm, self.gvec, self.nlocs, self.locs, self.vecs))
        self.access   = None



def cg(A, b, x, imax=50, eps=1e-6):
    """
    A, b, x  : matrix, rhs, solution
    imax     : maximum allowed iterations
    eps      : tolerance for convergence
    """
    # allocate work vectors
    r = b.duplicate()
    d = b.duplicate()
    q = b.duplicate()
    # initialization
    i = 0
    A.mult(x, r)
    r.aypx(-1, b)
    r.copy(d)
    delta_0 = r.dot(r)
    delta = delta_0
    # enter iteration loop
    while i < imax and \
          delta > delta_0 * eps**2:
        A.mult(d, q)
        alpha = delta / d.dot(q)
        x.axpy(+alpha, d)
        r.axpy(-alpha, q)
        delta_old = delta
        delta = r.dot(r)
        beta = delta / delta_old
        d.aypx(beta, r)
        i = i + 1
    return i, delta**0.5



class RegressorType(object):
    """REGRESSOR solver type.

    See Also
    --------
    petsc.PetscRegressorType

    """
    LINEAR = S_(PETSCREGRESSORLINEAR)


class RegressorLinearType(object):
    """Linear regressor type.

    See Also
    --------
    petsc.PetscRegressorLinearType
    """
    OLS   = REGRESSOR_LINEAR_OLS
    LASSO = REGRESSOR_LINEAR_LASSO
    RIDGE = REGRESSOR_LINEAR_RIDGE


cdef class Regressor(Object):
    """Regression solver.

    REGRESSOR  is described in the `PETSc manual <petsc:manual/regressor>`.

    See Also
    --------
    petsc.PetscRegressor

    """

    Type = RegressorType
    LinearType = RegressorLinearType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.regressor
        self.regressor = NULL

    def view(self, Viewer viewer=None) -> None:
        """View the solver.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        petsc.PetscRegressorView

        """
        cdef PetscViewer cviewer = NULL
        if viewer is not None: cviewer = viewer.vwr
        CHKERR(PetscRegressorView(self.regressor, cviewer))

    def create(self, comm=None) -> Self:
        """Create a REGRESSOR solver.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        Sys.getDefaultComm, petsc.PetscRegressorCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscRegressor newregressor = NULL
        CHKERR(PetscRegressorCreate(ccomm, &newregressor))
        PetscCLEAR(self.obj); self.regressor = newregressor
        return self

    def setRegularizerWeight(self, weight: float) -> None:
        """Set the weight to be used for the regularizer.

        Logically collective.

        See Also
        --------
        setType, petsc.PetscRegressorSetRegularizerWeight
        """
        CHKERR(PetscRegressorSetRegularizerWeight(self.regressor, weight))

    def setFromOptions(self) -> None:
        """Configure the solver from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.PetscRegressorSetFromOptions
        """
        CHKERR(PetscRegressorSetFromOptions(self.regressor))

    def setUp(self) -> None:
        """Set up the internal data structures for using the solver.

        Collective.

        See Also
        --------
        petsc.PetscRegressorSetUp

        """
        CHKERR(PetscRegressorSetUp(self.regressor))

    def fit(self, Mat X, Vec y) -> None:
        """Fit the regression problem.

        Collective.

        Parameters
        ----------
        X
            The matrix of training data
        y
            The vector of target values from the training dataset

        See Also
        --------
        petsc.PetscRegressorPredict

        """
        CHKERR(PetscRegressorFit(self.regressor, X.mat, y.vec))

    def predict(self, Mat X, Vec y) -> None:
        """Predict the regression problem.

        Collective.

        Parameters
        ----------
        X
            The matrix of unlabeled observations
        y
            The vector of predicted labels

        See Also
        --------
        petsc.PetscRegressorFit

        """
        CHKERR(PetscRegressorPredict(self.regressor, X.mat, y.vec))

    def getTAO(self) -> TAO:
        """Return the underlying `TAO` object .

        Not collective.

        See Also
        --------
        getLinearKSP, petsc.PetscRegressorGetTao
        """
        cdef TAO tao = TAO()
        CHKERR(PetscRegressorGetTao(self.regressor, &tao.tao))
        CHKERR(PetscINCREF(tao.obj))
        return tao

    def reset(self) -> None:
        """Destroy internal data structures of the solver.

        Collective.

        See Also
        --------
        petsc.PetscRegressorDestroy

        """
        CHKERR(PetscRegressorReset(self.regressor))

    def destroy(self) -> Self:
        """Destroy the solver.

        Collective.

        See Also
        --------
        petsc.PetscRegressorDestroy

        """
        CHKERR(PetscRegressorDestroy(&self.regressor))
        return self

    def setType(self, regressor_type: Type | str) -> None:
        """Set the type of the solver.

        Logically collective.

        Parameters
        ----------
        regressor_type
            The type of the solver.

        See Also
        --------
        getType, petsc.PetscRegressorSetType

        """
        cdef PetscRegressorType cval = NULL
        regressor_type = str2bytes(regressor_type, &cval)
        CHKERR(PetscRegressorSetType(self.regressor, cval))

    def getType(self) -> str:
        """Return the type of the solver.

        Not collective.

        See Also
        --------
        setType, petsc.PetscRegressorGetType

        """
        cdef PetscRegressorType ctype = NULL
        CHKERR(PetscRegressorGetType(self.regressor, &ctype))
        return bytes2str(ctype)

    # --- Linear ---

    def setLinearFitIntercept(self, flag: bool) -> None:
        """Set a flag to indicate that the intercept should be calculated.

        Logically collective.

        See Also
        --------
        petsc.PetscRegressorLinearSetFitIntercept
        """
        cdef PetscBool fitintercept = flag
        CHKERR(PetscRegressorLinearSetFitIntercept(self.regressor, fitintercept))

    def setLinearUseKSP(self, flag: bool) -> None:
        """Set a flag to indicate that `KSP` instead of `TAO` solvers should be used.

        Logically collective.

        See Also
        --------
        petsc.PetscRegressorLinearSetUseKSP
        """
        cdef PetscBool useksp = flag
        CHKERR(PetscRegressorLinearSetUseKSP(self.regressor, useksp))

    def getLinearKSP(self) -> KSP:
        """Returns the `KSP` context used by the linear regressor.

        Not collective.

        See Also
        --------
        petsc.PetscRegressorLinearGetKSP
        """
        cdef KSP ksp = KSP()
        CHKERR(PetscRegressorLinearGetKSP(self.regressor, &ksp.ksp))
        CHKERR(PetscINCREF(ksp.obj))
        return ksp

    def getLinearCoefficients(self) -> Vec:
        """Get a vector of the fitted coefficients from a linear regression model.

        Not collective.

        See Also
        --------
        getLinearIntercept, petsc.PetscRegressorLinearGetCoefficients
        """
        cdef Vec coeffs = Vec()
        CHKERR(PetscRegressorLinearGetCoefficients(self.regressor, &coeffs.vec))
        CHKERR(PetscINCREF(coeffs.obj))
        return coeffs

    def getLinearIntercept(self) -> Scalar:
        """Get the intercept from a linear regression model.

        Not collective.

        See Also
        --------
        setLinearFitIntercept, petsc.PetscRegressorLinearGetIntercept
        """
        cdef PetscScalar intercept = 0.0
        CHKERR(PetscRegressorLinearGetIntercept(self.regressor, &intercept))
        return toScalar(intercept)

    def setLinearType(self, lineartype: RegressorLinearType) -> None:
        """Set the type of linear regression to be performed.

        Logically collective.

        See Also
        --------
        getLinearType, petsc.PetscRegressorLinearSetType
        """
        CHKERR(PetscRegressorLinearSetType(self.regressor, lineartype))

    def getLinearType(self) -> RegressorLinearType:
        """Return the type of the linear regressor.

        Not collective.

        See Also
        --------
        setLinearType, petsc.PetscRegressorLinearGetType
        """
        cdef PetscRegressorLinearType cval = REGRESSOR_LINEAR_OLS
        CHKERR(PetscRegressorLinearGetType(self.regressor, &cval))
        return cval

del RegressorType



import unittest
from petsc4py import PETSc

# --------------------------------------------------------------------


class MyODE:
    """
    du/dt + u**2 = 0;
    u0,u1,u2 = 1,2,3
    """

    def __init__(self):
        self.rhsfunction_calls = 0
        self.rhsjacobian_calls = 0
        self.ifunction_calls = 0
        self.ijacobian_calls = 0
        self.presolve_calls = 0
        self.update_calls = 0
        self.postsolve_calls = 0
        self.monitor_calls = 0

    def rhsfunction(self, ts, t, u, F):
        # print ('MyODE.rhsfunction()')
        self.rhsfunction_calls += 1
        f = -(u * u)
        f.copy(F)

    def rhsjacobian(self, ts, t, u, J, P):
        # print ('MyODE.rhsjacobian()')
        self.rhsjacobian_calls += 1
        P.zeroEntries()
        diag = -2 * u
        P.setDiagonal(diag)
        P.assemble()
        if J != P:
            J.assemble()
        return True  # same_nz

    def ifunction(self, ts, t, u, du, F):
        # print ('MyODE.ifunction()')
        self.ifunction_calls += 1
        f = du + u * u
        f.copy(F)

    def ijacobian(self, ts, t, u, du, a, J, P):
        # print ('MyODE.ijacobian()')
        self.ijacobian_calls += 1
        P.zeroEntries()
        diag = a + 2 * u
        P.setDiagonal(diag)
        P.assemble()
        if J != P:
            J.assemble()
        return True  # same_nz

    def monitor(self, ts, s, t, u):
        self.monitor_calls += 1
        # dt = ts.time_step
        # ut = ts.vec_sol.norm()
        # prn = PETSc.Sys.Print
        # prn('TS: step %2d, T:%f, dT:%f, u:%f' % (s,t,dt,ut))


class BaseTestTSNonlinear:
    TYPE = None

    def setUp(self):
        self.ts = PETSc.TS().create(PETSc.COMM_SELF)
        eft = PETSc.TS.ExactFinalTime.STEPOVER
        self.ts.setExactFinalTime(eft)
        ptype = PETSc.TS.ProblemType.NONLINEAR
        self.ts.setProblemType(ptype)
        self.ts.setType(self.TYPE)
        if PETSc.ScalarType().dtype.char in 'fF':
            snes = self.ts.getSNES()
            snes.setTolerances(rtol=1e-6)

    def tearDown(self):
        self.ts = None
        PETSc.garbage_cleanup()


class BaseTestTSNonlinearRHS(BaseTestTSNonlinear):
    def testSolveRHS(self, nullsol=False):
        ts = self.ts
        dct = self.ts.getDict()
        self.assertTrue(dct is not None)
        self.assertTrue(isinstance(dct, dict))

        ode = MyODE()
        J = PETSc.Mat().create(ts.comm)
        J.setSizes(3)
        J.setFromOptions()
        J.setUp()
        u, f = J.createVecs()

        ts.setAppCtx(ode)
        ts.setRHSFunction(ode.rhsfunction, f)
        ts.setRHSJacobian(ode.rhsjacobian, J, J)
        ts.setMonitor(ode.monitor)

        ts.snes.ksp.pc.setType('none')

        T0, dT, nT = 0.00, 0.1, 10
        T = T0 + nT * dT
        ts.setTime(T0)
        ts.setTimeStep(dT)
        ts.setMaxTime(T)
        ts.setMaxSteps(nT)
        ts.setFromOptions()
        u[0], u[1], u[2] = 1, 2, 3
        if nullsol:
            ts.setSolution(u)
            ts.solve()
        else:
            ts.solve(u)

        self.assertTrue(ode.rhsfunction_calls > 0)
        self.assertTrue(ode.rhsjacobian_calls > 0)

        dct = self.ts.getDict()
        self.assertTrue('__appctx__' in dct)
        self.assertTrue('__rhsfunction__' in dct)
        self.assertTrue('__rhsjacobian__' in dct)
        self.assertTrue('__monitor__' in dct)

        n = ode.monitor_calls
        ts.monitor(ts.step_number, ts.time)
        self.assertEqual(ode.monitor_calls, n + 1)
        n = ode.monitor_calls
        ts.monitorCancel()
        ts.monitor(ts.step_number, ts.time)
        self.assertEqual(ode.monitor_calls, n)

    def testFDColorRHS(self):
        ts = self.ts
        ode = MyODE()
        J = PETSc.Mat().create(ts.comm)
        J.setSizes(5)
        J.setType('aij')
        J.setPreallocationNNZ(nnz=1)
        u, f = J.createVecs()

        ts.setAppCtx(ode)
        ts.setRHSFunction(ode.rhsfunction, f)
        ts.setRHSJacobian(ode.rhsjacobian, J, J)
        ts.setMonitor(ode.monitor)

        T0, dT, nT = 0.00, 0.1, 10
        T = T0 + nT * dT
        ts.setTime(T0)
        ts.setTimeStep(dT)
        ts.setMaxTime(T)
        ts.setMaxSteps(nT)
        ts.setFromOptions()
        u[0], u[1], u[2] = 1, 2, 3

        ts.setSolution(u)
        ode.rhsjacobian(ts, 0, u, J, J)
        ts.setUp()
        ts.snes.setUseFD(True)
        ts.solve(u)

    def testResetAndSolveRHS(self):
        self.ts.reset()
        self.ts.setStepNumber(0)
        self.testSolveRHS()
        self.ts.reset()
        self.ts.setStepNumber(0)
        self.testSolveRHS()
        self.ts.reset()
        self.ts.setStepNumber(0)
        self.testSolveRHS(nullsol=True)
        self.ts.reset()


class BaseTestTSNonlinearI(BaseTestTSNonlinear):
    def testSolveI(self):
        ts = self.ts
        dct = self.ts.getDict()
        self.assertTrue(dct is not None)
        self.assertTrue(isinstance(dct, dict))

        ode = MyODE()
        J = PETSc.Mat().create(ts.comm)
        J.setSizes(3)
        J.setFromOptions()
        J.setUp()
        u, f = J.createVecs()

        ts.setAppCtx(ode)
        ts.setIFunction(ode.ifunction, f)
        ts.setIJacobian(ode.ijacobian, J, J)
        ts.setMonitor(ode.monitor)

        ts.snes.ksp.pc.setType('none')

        T0, dT, nT = 0.00, 0.1, 10
        T = T0 + nT * dT
        ts.setTime(T0)
        ts.setTimeStep(dT)
        ts.setMaxTime(T)
        ts.setMaxSteps(nT)
        ts.setFromOptions()
        u[0], u[1], u[2] = 1, 2, 3
        ts.solve(u)

        self.assertTrue(ode.ifunction_calls > 0)
        self.assertTrue(ode.ijacobian_calls > 0)

        dct = self.ts.getDict()
        self.assertTrue('__appctx__' in dct)
        self.assertTrue('__ifunction__' in dct)
        self.assertTrue('__ijacobian__' in dct)
        self.assertTrue('__monitor__' in dct)

        n = ode.monitor_calls
        ts.monitor(ts.step_number, ts.time)
        self.assertEqual(ode.monitor_calls, n + 1)
        n = ode.monitor_calls
        ts.monitorCancel()
        ts.monitor(ts.step_number, ts.time)
        self.assertEqual(ode.monitor_calls, n)

    def testFDColorI(self):
        ts = self.ts
        ode = MyODE()
        J = PETSc.Mat().create(ts.comm)
        J.setSizes(5)
        J.setType('aij')
        J.setPreallocationNNZ(nnz=1)
        J.setFromOptions()
        u, f = J.createVecs()

        ts.setAppCtx(ode)
        ts.setIFunction(ode.ifunction, f)
        ts.setIJacobian(ode.ijacobian, J, J)
        ts.setMonitor(ode.monitor)

        T0, dT, nT = 0.00, 0.1, 10
        T = T0 + nT * dT
        ts.setTime(T0)
        ts.setTimeStep(dT)
        ts.setMaxTime(T)
        ts.setMaxSteps(nT)
        ts.setFromOptions()
        u[0], u[1], u[2] = 1, 2, 3

        ts.setSolution(u)
        ode.ijacobian(ts, 0, u, 0 * u, 1, J, J)
        ts.setUp()
        ts.snes.setUseFD(True)
        ts.solve(u)

    def testResetAndSolveI(self):
        self.ts.reset()
        self.ts.setStepNumber(0)
        self.testSolveI()
        self.ts.reset()
        self.ts.setStepNumber(0)
        self.testSolveI()
        self.ts.reset()


class TestTSBeuler(BaseTestTSNonlinearRHS, BaseTestTSNonlinearI, unittest.TestCase):
    TYPE = PETSc.TS.Type.BEULER


class TestTSCN(BaseTestTSNonlinearRHS, BaseTestTSNonlinearI, unittest.TestCase):
    TYPE = PETSc.TS.Type.CN


class TestTSTheta(BaseTestTSNonlinearRHS, BaseTestTSNonlinearI, unittest.TestCase):
    TYPE = PETSc.TS.Type.THETA


class TestTSAlpha(BaseTestTSNonlinearRHS, BaseTestTSNonlinearI, unittest.TestCase):
    TYPE = PETSc.TS.Type.ALPHA


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# Simple linear regression test
# =============================
#
# Use the petsc4py interface to PetscRegressor to solve the same quadratic
# polynomial data-fitting problem that is solved in src/ml/regressor/tests/ex1.c.

import sys
import petsc4py

petsc4py.init(sys.argv)

from petsc4py import PETSc

comm = PETSc.COMM_WORLD
size = comm.getSize()
rank = comm.getRank()

OptDB = PETSc.Options()

y_array = [0.2, -0.3, -0.8, -0.3, 1.2]
X_array = [-1.0, 0.5, -0.5, -0.25, 0.0, -0.5, 0.5, -0.25, 1.0, 0.5]
rows_ix = [0, 1, 2, 3, 4]
cols_ix = [0, 1]

X = PETSc.Mat().create(comm=comm)
X.setSizes((5,2))
X.setFromOptions()
X.setUp()

y = PETSc.Vec().create(comm=comm)
y.setSizes(5)
y.setFromOptions()

if not rank :
    X.setValues(rows_ix,cols_ix,X_array,addv=True)
    y.setValues(rows_ix,y_array,addv=False)

X.assemblyBegin(X.AssemblyType.FINAL)
X.assemblyEnd(X.AssemblyType.FINAL)
y.assemblyBegin()
y.assemblyEnd()

y_predicted = y.duplicate()

regressor = PETSc.Regressor().create(comm=comm)
regressor.setType(PETSc.Regressor.Type.LINEAR)
regressor.fit(X,y)
regressor.predict(X,y_predicted)
y_predicted.view()






# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
import os
import sys
import optparse
import unittest

__unittest = True

components = [
    'PETSc',
]


def getoptionparser():
    parser = optparse.OptionParser()

    parser.add_option(
        '-q',
        '--quiet',
        action='store_const',
        const=0,
        dest='verbose',
        default=1,
        help='do not print status messages to stdout',
    )
    parser.add_option(
        '-v',
        '--verbose',
        action='store_const',
        const=2,
        dest='verbose',
        default=1,
        help='print status messages to stdout',
    )
    parser.add_option(
        '-i',
        '--include',
        type='string',
        action='append',
        dest='include',
        default=[],
        help='include tests matching PATTERN',
        metavar='PATTERN',
    )
    parser.add_option(
        '-e',
        '--exclude',
        type='string',
        action='append',
        dest='exclude',
        default=[],
        help='exclude tests matching PATTERN',
        metavar='PATTERN',
    )
    parser.add_option(
        '-k',
        '--pattern',
        type='string',
        action='append',
        dest='patterns',
        default=[],
        help='only run tests which match the given substring',
    )
    parser.add_option(
        '-f',
        '--failfast',
        action='store_true',
        dest='failfast',
        default=False,
        help='Stop on first failure',
    )
    parser.add_option(
        '--no-builddir',
        action='store_false',
        dest='builddir',
        default=True,
        help='disable testing from build directory',
    )
    parser.add_option(
        '--path',
        type='string',
        action='append',
        dest='path',
        default=[],
        help='prepend PATH to sys.path',
        metavar='PATH',
    )
    parser.add_option(
        '--arch',
        type='string',
        action='store',
        dest='arch',
        default=None,
        help='use PETSC_ARCH',
        metavar='PETSC_ARCH',
    )
    parser.add_option(
        '-s',
        '--summary',
        action='store_true',
        dest='summary',
        default=0,
        help='print PETSc log summary',
    )
    parser.add_option(
        '--no-memdebug',
        action='store_false',
        dest='memdebug',
        default=True,
        help='Do not use PETSc memory debugging',
    )
    return parser


def getbuilddir():
    try:
        try:
            from setuptools.dist import Distribution
        except ImportError:
            from distutils.dist import Distribution
        try:
            from setuptools.command.build import build
        except ImportError:
            from distutils.command.build import build
        cmd_obj = build(Distribution())
        cmd_obj.finalize_options()
        return cmd_obj.build_platlib
    except Exception:
        return None


def getprocessorinfo():
    try:
        name = os.uname()[1]
    except Exception:
        import platform

        name = platform.uname()[1]
    from petsc4py.PETSc import COMM_WORLD

    rank = COMM_WORLD.getRank()
    return (rank, name)


def getlibraryinfo(name):
    modname = f'{name.lower()}4py.{name}'
    module = __import__(modname, fromlist=[name])
    (major, minor, micro), devel = module.Sys.getVersion(devel=True)
    r = not devel
    if r:
        release = 'release'
    else:
        release = 'development'
    arch = module.__arch__
    return "%s %d.%d.%d %s (conf: '%s')" % (name, major, minor, micro, release, arch)


def getpythoninfo():
    x, y, z = sys.version_info[:3]
    return 'Python %d.%d.%d (%s)' % (x, y, z, sys.executable)


def getpackageinfo(pkg):
    try:
        pkg = __import__(pkg)
    except ImportError:
        return None
    name = pkg.__name__
    version = pkg.__version__
    path = pkg.__path__[0]
    return f'{name} {version} ({path})'


def setup_python(options):
    rootdir = os.path.dirname(os.path.dirname(__file__))
    builddir = getbuilddir()
    if builddir is not None:
        builddir = os.path.join(rootdir, builddir)
    if options.builddir and builddir is not None and os.path.exists(builddir):
        sys.path.insert(0, builddir)
    if options.path:
        path = options.path[:]
        path.reverse()
        for p in path:
            sys.path.insert(0, p)


def setup_unittest(options):
    try:
        from unittest.runner import _WritelnDecorator
    except ImportError:
        from unittest import _WritelnDecorator
    #
    writeln_orig = _WritelnDecorator.writeln

    def writeln(self, message=''):
        try:
            self.stream.flush()
        except Exception:
            pass
        writeln_orig(self, message)
        try:
            self.stream.flush()
        except Exception:
            pass

    _WritelnDecorator.writeln = writeln


def import_package(options, pkgname):
    args = [sys.argv[0]]
    if options.memdebug:
        args.append('-malloc_debug')
        args.append('-malloc_dump')
    if options.summary:
        args.append('-log_view')
    package = __import__(pkgname)
    package.init(args, arch=options.arch)


def print_banner(options):
    r, n = getprocessorinfo()
    prefix = '[%d@%s]' % (r, n)

    def writeln(message='', endl='\n'):
        if message is None:
            return
        from petsc4py.PETSc import Sys

        message = f'{prefix} {message}'
        Sys.syncPrint(message, endl=endl, flush=True)

    if options.verbose:
        writeln(getpythoninfo())
        writeln(getpackageinfo('numpy'))
        for entry in components:
            writeln(getlibraryinfo(entry))
            writeln(getpackageinfo('%s4py' % entry.lower()))


def load_tests(options, args):
    from glob import glob
    import re

    testsuitedir = os.path.dirname(__file__)
    sys.path.insert(0, testsuitedir)
    pattern = 'test_*.py'
    wildcard = os.path.join(testsuitedir, pattern)
    testfiles = glob(wildcard)
    testfiles.sort()
    testsuite = unittest.TestSuite()
    testloader = unittest.TestLoader()
    if options.patterns:
        testloader.testNamePatterns = [ # novermin
            ('*%s*' % p) if ('*' not in p) else p for p in options.patterns
        ]
    include = exclude = None
    if options.include:
        include = re.compile('|'.join(options.include)).search
    if options.exclude:
        exclude = re.compile('|'.join(options.exclude)).search
    for testfile in testfiles:
        filename = os.path.basename(testfile)
        testname = os.path.splitext(filename)[0]
        if (exclude and exclude(testname)) or (include and not include(testname)):
            continue
        module = __import__(testname)
        for arg in args:
            try:
                cases = testloader.loadTestsFromNames((arg,), module)
                testsuite.addTests(cases)
            except AttributeError:
                pass
        if not args:
            cases = testloader.loadTestsFromModule(module)
            testsuite.addTests(cases)
    return testsuite


def run_tests(options, testsuite, runner=None):
    if runner is None:
        runner = unittest.TextTestRunner(verbosity=options.verbose)
        runner.failfast = options.failfast
    result = runner.run(testsuite)
    return result.wasSuccessful()


def abort(code=1):
    os.abort()


def shutdown(success):
    pass


def main(args=None):
    pkgname = '%s4py' % components[-1].lower()
    parser = getoptionparser()
    (options, args) = parser.parse_args(args)
    setup_python(options)
    setup_unittest(options)
    import_package(options, pkgname)
    print_banner(options)
    testsuite = load_tests(options, args)
    success = run_tests(options, testsuite)
    if not success and options.failfast:
        abort()
    shutdown(success)
    return not success


if __name__ == '__main__':
    import sys

    sys.dont_write_bytecode = True
    sys.exit(main())
